#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.omps_edr module."""

import os
import unittest
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        file_content = {}
        attrs = []
        if "SO2NRT" in filename:
            k = "HDFEOS/SWATHS/OMPS Column Amount SO2/Data Fields/ColumnAmountSO2_TRM"
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/ScaleFactor"] = 1.1
            file_content[k + "/attr/Offset"] = 0.1
            file_content[k + "/attr/MissingValue"] = -1
            file_content[k + "/attr/Title"] = "Vertical Column Amount SO2 (TRM)"
            file_content[k + "/attr/Units"] = "D.U."
            file_content[k + "/attr/ValidRange"] = (-10, 2000)
            k = "HDFEOS/SWATHS/OMPS Column Amount SO2/Geolocation Fields/Longitude"
            file_content[k] = DEFAULT_LON_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/ScaleFactor"] = 1.1
            file_content[k + "/attr/Offset"] = 0.1
            file_content[k + "/attr/Units"] = "deg"
            file_content[k + "/attr/MissingValue"] = -1
            file_content[k + "/attr/Title"] = "Geodetic Longitude"
            file_content[k + "/attr/ValidRange"] = (-180, 180)
            k = "HDFEOS/SWATHS/OMPS Column Amount SO2/Geolocation Fields/Latitude"
            file_content[k] = DEFAULT_LAT_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/ScaleFactor"] = 1.1
            file_content[k + "/attr/Offset"] = 0.1
            file_content[k + "/attr/Units"] = "deg"
            file_content[k + "/attr/MissingValue"] = -1
            file_content[k + "/attr/Title"] = "Geodetic Latitude"
            file_content[k + "/attr/ValidRange"] = (-90, 90)
        elif "NMSO2" in filename:
            file_content["GEOLOCATION_DATA/Longitude"] = DEFAULT_LON_DATA
            file_content["GEOLOCATION_DATA/Longitude/shape"] = DEFAULT_FILE_SHAPE
            file_content["GEOLOCATION_DATA/Longitude/attr/valid_max"] = 180
            file_content["GEOLOCATION_DATA/Longitude/attr/valid_min"] = -180
            file_content["GEOLOCATION_DATA/Longitude/attr/_FillValue"] = -1.26765e+30
            file_content["GEOLOCATION_DATA/Longitude/attr/long_name"] = "Longitude"
            file_content["GEOLOCATION_DATA/Longitude/attr/standard_name"] = "longitude"
            file_content["GEOLOCATION_DATA/Longitude/attr/units"] = "degrees_east"
            file_content["GEOLOCATION_DATA/Latitude"] = DEFAULT_LAT_DATA
            file_content["GEOLOCATION_DATA/Latitude/shape"] = DEFAULT_FILE_SHAPE
            file_content["GEOLOCATION_DATA/Latitude/attr/valid_max"] = 90
            file_content["GEOLOCATION_DATA/Latitude/attr/valid_min"] = -90
            file_content["GEOLOCATION_DATA/Latitude/attr/_FillValue"] = -1.26765e+30
            file_content["GEOLOCATION_DATA/Latitude/attr/long_name"] = "Latitude"
            file_content["GEOLOCATION_DATA/Latitude/attr/standard_name"] = "latitude"
            file_content["GEOLOCATION_DATA/Latitude/attr/units"] = "degress_north"

            k = "SCIENCE_DATA/ColumnAmountSO2_TRM"
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/_FillValue"] = -1.26765e+30
            file_content[k + "/attr/long_name"] = "Column Amount SO2 (TRM)"
            file_content[k + "/attr/units"] = "DU"
            file_content[k + "/attr/valid_max"] = 2000
            file_content[k + "/attr/valid_min"] = -10

            k = "SCIENCE_DATA/ColumnAmountSO2_STL"
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/_FillValue"] = -1.26765e+30
            file_content[k + "/attr/long_name"] = "Column Amount SO2 (STL)"
            file_content[k + "/attr/units"] = "DU"

            k = "SCIENCE_DATA/ColumnAmountSO2_TRL"
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/_FillValue"] = -1.26765e+30
            file_content[k + "/attr/long_name"] = "Column Amount SO2 (TRL)"
            file_content[k + "/attr/units"] = "DU"
            file_content[k + "/attr/valid_max"] = 2000
            file_content[k + "/attr/valid_min"] = -10
            file_content[k + "/attr/DIMENSION_LIST"] = [10, 10]
            attrs = ["_FillValue", "long_name", "units", "valid_max", "valid_min", "DIMENSION_LIST"]

            k = "SCIENCE_DATA/ColumnAmountSO2_TRU"
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/long_name"] = "Column Amount SO2 (TRU)"
            file_content[k + "/attr/units"] = "DU"
            file_content[k + "/attr/valid_max"] = 2000
            file_content[k + "/attr/valid_min"] = -10

            # Dataset with out unit
            k = "SCIENCE_DATA/ColumnAmountSO2_PBL"
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/_FillValue"] = -1.26765e+30
            file_content[k + "/attr/long_name"] = "Column Amount SO2 (PBL)"
            file_content[k + "/attr/valid_max"] = 2000
            file_content[k + "/attr/valid_min"] = -10
        else:
            for k in ["Reflectivity331", "UVAerosolIndex"]:
                k = "SCIENCE_DATA/" + k
                file_content[k] = DEFAULT_FILE_DATA
                file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
                file_content[k + "/attr/Units"] = "Unitless"
                if k == "UVAerosolIndex":
                    file_content[k + "/attr/ValidRange"] = (-30, 30)
                    file_content[k + "/attr/Title"] = "UV Aerosol Index"
                else:
                    file_content[k + "/attr/ValidRange"] = (-0.15, 1.15)
                    file_content[k + "/attr/Title"] = "Effective Surface Reflectivity at 331 nm"
                file_content[k + "/attr/_FillValue"] = -1.
            file_content["GEOLOCATION_DATA/Longitude"] = DEFAULT_LON_DATA
            file_content["GEOLOCATION_DATA/Longitude/shape"] = DEFAULT_FILE_SHAPE
            file_content["GEOLOCATION_DATA/Longitude/attr/ValidRange"] = (-180, 180)
            file_content["GEOLOCATION_DATA/Longitude/attr/_FillValue"] = -999.
            file_content["GEOLOCATION_DATA/Longitude/attr/Title"] = "Geodetic Longitude"
            file_content["GEOLOCATION_DATA/Longitude/attr/Units"] = "deg"
            file_content["GEOLOCATION_DATA/Latitude"] = DEFAULT_LAT_DATA
            file_content["GEOLOCATION_DATA/Latitude/shape"] = DEFAULT_FILE_SHAPE
            file_content["GEOLOCATION_DATA/Latitude/attr/ValidRange"] = (-90, 90)
            file_content["GEOLOCATION_DATA/Latitude/attr/_FillValue"] = -999.
            file_content["GEOLOCATION_DATA/Latitude/attr/Title"] = "Geodetic Latitude"
            file_content["GEOLOCATION_DATA/Latitude/attr/Units"] = "deg"

        convert_file_content_to_data_array(file_content, attrs)
        return file_content


class TestOMPSEDRReader(unittest.TestCase):
    """Test OMPS EDR Reader."""

    yaml_file = "omps_edr.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.omps_edr import EDREOSFileHandler, EDRFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(EDRFileHandler, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True
        self.p2 = mock.patch.object(EDREOSFileHandler, "__bases__", (EDRFileHandler,))
        self.fake_handler2 = self.p2.start()
        self.p2.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p2.stop()
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "OMPS-NPP-TC_EDR_SO2NRT-2016m0607t192031-o00001-2016m0607t192947.he5",
            "OMPS-NPP-TC_EDR_TO3-v1.0-2016m0607t192031-o00001-2016m0607t192947.h5",
            "OMPS-NPP_NMSO2-PCA-L2_v1.1_2018m1129t112824_o00001_2018m1129t114426.h5",
        ])
        assert len(loadables) == 3
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_basic_load_so2(self):
        """Test basic load of so2 datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "OMPS-NPP-TC_EDR_SO2NRT-2016m0607t192031-o00001-2016m0607t192947.he5",
            "OMPS-NPP-TC_EDR_TO3-v1.0-2016m0607t192031-o00001-2016m0607t192947.h5",
            "OMPS-NPP_NMSO2-PCA-L2_v1.1_2018m1129t112824_o00001_2018m1129t114426.h5",
        ])
        assert len(loadables) == 3
        r.create_filehandlers(loadables)
        ds = r.load(["so2_trm"])
        assert len(ds) == 1
        for d in ds.values():
            assert d.attrs["resolution"] == 50000
            assert d.shape == DEFAULT_FILE_SHAPE
            assert "area" in d.attrs
            assert d.attrs["area"] is not None

        ds = r.load(["tcso2_trm_sampo"])
        assert len(ds) == 1
        for d in ds.values():
            assert d.attrs["resolution"] == 50000
            assert d.shape == DEFAULT_FILE_SHAPE

        ds = r.load(["tcso2_stl_sampo"])
        assert len(ds) == 0

        # Dataset without _FillValue
        ds = r.load(["tcso2_tru_sampo"])
        assert len(ds) == 1

        # Dataset without unit
        ds = r.load(["tcso2_pbl_sampo"])
        assert len(ds) == 0

    def test_basic_load_to3(self):
        """Test basic load of to3 datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "OMPS-NPP-TC_EDR_SO2NRT-2016m0607t192031-o00001-2016m0607t192947.he5",
            "OMPS-NPP-TC_EDR_TO3-v1.0-2016m0607t192031-o00001-2016m0607t192947.h5",
            "OMPS-NPP_NMSO2-PCA-L2_v1.1_2018m1129t112824_o00001_2018m1129t114426.h5",
        ])
        assert len(loadables) == 3
        r.create_filehandlers(loadables)
        ds = r.load(["reflectivity_331", "uvaerosol_index"])
        assert len(ds) == 2
        for d in ds.values():
            assert d.attrs["resolution"] == 50000
            assert d.shape == DEFAULT_FILE_SHAPE
            assert "area" in d.attrs
            assert d.attrs["area"] is not None

    @mock.patch("satpy.readers.core.hdf5.HDF5FileHandler._get_reference")
    @mock.patch("h5py.File")
    def test_load_so2_DIMENSION_LIST(self, mock_h5py_file, mock_hdf5_utils_get_reference):
        """Test load of so2 datasets with DIMENSION_LIST."""
        from satpy.readers.core.loading import load_reader
        mock_h5py_file.return_value = mock.MagicMock()
        mock_hdf5_utils_get_reference.return_value = [[[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]]]
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "OMPS-NPP_NMSO2-PCA-L2_v1.1_2018m1129t112824_o00001_2018m1129t114426.h5",
        ])
        r.create_filehandlers(loadables)

        ds = r.load(["tcso2_trl_sampo"])
        assert len(ds) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for MODIS L2 HDF reader."""

from __future__ import annotations

import dask
import dask.array as da
import numpy as np
import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy import Scene, available_readers
from satpy.tests.utils import CustomScheduler, make_dataid

from ._modis_fixtures import _shape_for_resolution

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - modis_l2_imapp_mask_byte1_file
# - modis_l2_imapp_mask_byte1_geo_files
# - modis_l2_imapp_snowmask_file
# - modis_l2_imapp_snowmask_geo_files
# - modis_l2_nasa_mod06_file
# - modis_l2_nasa_mod35_file
# - modis_l2_nasa_mod99_file
# - modis_l2_nasa_mod35_mod03_files


def _check_shared_metadata(data_arr, expect_area=False):
    assert data_arr.attrs["sensor"] == "modis"
    assert data_arr.attrs["platform_name"] == "Terra"
    assert "rows_per_scan" in data_arr.attrs
    assert isinstance(data_arr.attrs["rows_per_scan"], int)
    assert data_arr.attrs["reader"] == "modis_l2"
    if expect_area:
        assert data_arr.attrs.get("area") is not None
    else:
        assert "area" not in data_arr.attrs


class TestModisL2:
    """Test MODIS L2 reader."""

    def test_available_reader(self):
        """Test that MODIS L2 reader is available."""
        assert "modis_l2" in available_readers()

    def test_scene_available_datasets(self, modis_l2_nasa_mod35_file):
        """Test that datasets are available."""
        scene = Scene(reader="modis_l2", filenames=modis_l2_nasa_mod35_file)
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert "cloud_mask" in available_datasets
        assert "latitude" in available_datasets
        assert "longitude" in available_datasets

    @pytest.mark.parametrize(
        ("input_files", "has_5km", "has_500", "has_250", "default_res"),
        [
            (lazy_fixture("modis_l2_nasa_mod35_file"),
             True, False, False, 1000),
        ]
    )
    def test_load_longitude_latitude(self, input_files, has_5km, has_500, has_250, default_res):
        """Test that longitude and latitude datasets are loaded correctly."""
        from .test_modis_l1b import _load_and_check_geolocation
        scene = Scene(reader="modis_l2", filenames=input_files)
        shape_5km = _shape_for_resolution(5000)
        shape_500m = _shape_for_resolution(500)
        shape_250m = _shape_for_resolution(250)
        default_shape = _shape_for_resolution(default_res)
        with dask.config.set(scheduler=CustomScheduler(max_computes=1 + has_5km + has_500 + has_250)):
            _load_and_check_geolocation(scene, "*", default_res, default_shape, True,
                                        check_callback=_check_shared_metadata)
            _load_and_check_geolocation(scene, 5000, 5000, shape_5km, has_5km,
                                        check_callback=_check_shared_metadata)
            _load_and_check_geolocation(scene, 500, 500, shape_500m, has_500,
                                        check_callback=_check_shared_metadata)
            _load_and_check_geolocation(scene, 250, 250, shape_250m, has_250,
                                        check_callback=_check_shared_metadata)

    def test_load_quality_assurance(self, modis_l2_nasa_mod35_file):
        """Test loading quality assurance."""
        scene = Scene(reader="modis_l2", filenames=modis_l2_nasa_mod35_file)
        dataset_name = "quality_assurance"
        scene.load([dataset_name])
        quality_assurance_id = make_dataid(name=dataset_name, resolution=1000)
        assert quality_assurance_id in scene
        quality_assurance = scene[quality_assurance_id]
        assert quality_assurance.shape == _shape_for_resolution(1000)
        _check_shared_metadata(quality_assurance, expect_area=True)

    @pytest.mark.parametrize(
        ("input_files", "loadables", "request_resolution", "exp_resolution", "exp_area"),
        [
            (lazy_fixture("modis_l2_nasa_mod35_mod03_files"),
             ["cloud_mask"],
             1000, 1000, True),
            (lazy_fixture("modis_l2_imapp_mask_byte1_geo_files"),
             ["cloud_mask", "land_sea_mask", "snow_ice_mask"],
             None, 1000, True),
        ]
    )
    def test_load_category_dataset(self, input_files, loadables, request_resolution, exp_resolution, exp_area):
        """Test loading category products."""
        scene = Scene(reader="modis_l2", filenames=input_files)
        kwargs = {"resolution": request_resolution} if request_resolution is not None else {}
        scene.load(loadables, **kwargs)
        for ds_name in loadables:
            cat_id = make_dataid(name=ds_name, resolution=exp_resolution)
            assert cat_id in scene
            cat_data_arr = scene[cat_id]
            assert isinstance(cat_data_arr.data, da.Array)
            cat_data_arr = cat_data_arr.compute()
            assert cat_data_arr.shape == _shape_for_resolution(exp_resolution)
            assert cat_data_arr.values[0, 0] == 0.0
            assert cat_data_arr.attrs.get("resolution") == exp_resolution
            # mask variables should be integers
            assert np.issubdtype(cat_data_arr.dtype, np.integer)
            assert cat_data_arr.attrs.get("_FillValue") is not None
            _check_shared_metadata(cat_data_arr, expect_area=exp_area)

    @pytest.mark.parametrize(
        ("input_files", "exp_area"),
        [
            (lazy_fixture("modis_l2_nasa_mod35_file"), False),
            (lazy_fixture("modis_l2_nasa_mod35_mod03_files"), True),
        ]
    )
    def test_load_250m_cloud_mask_dataset(self, input_files, exp_area):
        """Test loading 250m cloud mask."""
        scene = Scene(reader="modis_l2", filenames=input_files)
        dataset_name = "cloud_mask"
        scene.load([dataset_name], resolution=250)
        cloud_mask_id = make_dataid(name=dataset_name, resolution=250)
        assert cloud_mask_id in scene
        cloud_mask = scene[cloud_mask_id]
        assert isinstance(cloud_mask.data, da.Array)
        cloud_mask = cloud_mask.compute()
        assert cloud_mask.shape == _shape_for_resolution(250)
        assert cloud_mask.values[0, 0] == 0.0
        # mask variables should be integers
        assert np.issubdtype(cloud_mask.dtype, np.integer)
        assert cloud_mask.attrs.get("_FillValue") is not None
        _check_shared_metadata(cloud_mask, expect_area=exp_area)

    @pytest.mark.parametrize(
        ("input_files", "loadables", "exp_resolution", "exp_area", "exp_value"),
        [
            (lazy_fixture("modis_l2_nasa_mod06_file"), ["surface_pressure", "non_yaml_configured_2D_var"],
             5000, True, 4.0),
            (lazy_fixture("modis_l2_nasa_mod99_file"), ["non_yaml_configured_2D_var"], 1000, True, 4.0),
            # snow mask is considered a category product, factor/offset ignored
            (lazy_fixture("modis_l2_imapp_snowmask_file"), ["snow_mask"], 1000, False, 1.0),
            (lazy_fixture("modis_l2_imapp_snowmask_geo_files"), ["snow_mask"], 1000, True, 1.0),
        ]
    )
    def test_load_l2_dataset(self, input_files, loadables, exp_resolution, exp_area, exp_value):
        """Load and check an L2 variable."""
        scene = Scene(reader="modis_l2", filenames=input_files)
        scene.load(loadables)
        for ds_name in loadables:
            assert ds_name in scene
            data_arr = scene[ds_name]
            assert isinstance(data_arr.data, da.Array)
            data_arr = data_arr.compute()
            assert data_arr.values[0, 0] == exp_value
            assert data_arr.shape == _shape_for_resolution(exp_resolution)
            assert data_arr.attrs.get("resolution") == exp_resolution
            _check_shared_metadata(data_arr, expect_area=exp_area)

    def test_scene_dynamic_available_datasets(self, modis_l2_nasa_mod06_file):
        """Test available datasets method to dynmically add non configured datasets."""
        import xarray as xr
        scene = Scene(reader="modis_l2", filenames=modis_l2_nasa_mod06_file)
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert "surface_pressure" in available_datasets
        # make sure configured datasets are added again
        assert available_datasets.count("surface_pressure") == 1
        assert "non_yaml_configured_2D_var" in available_datasets
        file_ds = xr.open_dataset(modis_l2_nasa_mod06_file[0], engine="netcdf4")
        assert "non_yaml_configured_3D_var" not in available_datasets and "non_yaml_configured_3D_var" in file_ds # noqa PT018
        assert "non_yaml_configured_3D_var" in file_ds



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Demo data download for SEVIRI HRIT files."""

import logging
import os.path

from satpy import config
from satpy.demo.utils import download_url

logger = logging.getLogger(__name__)

ZENODO_BASE_URL = "https://zenodo.org/api/files/dcc5ab29-d8a3-4fb5-ab2b-adc405d18c23/"
FILENAME = "H-000-MSG4__-MSG4________-{channel:_<9s}-{segment:_<9s}-201802281500-__"


def download_seviri_hrit_20180228_1500(base_dir=None, subset=None):
    """Download the SEVIRI HRIT files for 2018-02-28T15:00.

    *subset* is a dictionary with the channels as keys and granules to download
    as values, eg::

      {"HRV": [1, 2, 3], "IR_108": [1, 2], "EPI": None}

    """
    files = generate_subset_of_filenames(subset)

    base_dir = base_dir or config.get("demo_data_dir", ".")
    subdir = os.path.join(base_dir, "seviri_hrit", "20180228_1500")
    os.makedirs(subdir, exist_ok=True)
    targets = []
    for the_file in files:
        target = os.path.join(subdir, the_file)
        targets.append(target)
        if os.path.isfile(target):
            continue
        download_url(ZENODO_BASE_URL + the_file, target)
    return targets


def generate_subset_of_filenames(subset=None, base_dir=""):
    """Generate SEVIRI HRIT filenames."""
    if subset is None:
        subset = _create_full_set()
    pattern = os.path.join(base_dir, FILENAME)
    files = []
    for channel, segments in subset.items():
        new_files = _generate_filenames(pattern, channel, segments)
        files.extend(new_files)
    return files


def _generate_filenames(pattern, channel, segments):
    """Generate the filenames for *channel* and *segments*."""
    if channel in ["PRO", "EPI"]:
        new_files = [pattern.format(channel="", segment=channel)]
    else:
        new_files = (pattern.format(channel=channel, segment=f"{segment:06d}") for segment in segments)
    return new_files


def _create_full_set():
    """Create the full set dictionary."""
    subset = {"HRV": range(1, 25),
              "EPI": None,
              "PRO": None}
    channels = ["IR_016", "IR_039", "IR_087", "IR_097", "IR_108", "IR_120", "IR_134",
                "VIS006", "VIS008",
                "WV_062", "WV_073"]
    for channel in channels:
        subset[channel] = range(1, 9)
    return subset



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.sar-c_safe module."""

import os
from datetime import datetime
from enum import Enum
from pathlib import Path

import numpy as np
import pytest
import yaml

geotiepoints = pytest.importorskip("geotiepoints", "1.7.5")

from satpy._config import PACKAGE_CONFIG_PATH  # noqa: E402
from satpy.dataset import DataQuery  # noqa: E402
from satpy.dataset.dataid import DataID  # noqa: E402
from satpy.readers.sar_c_safe import Calibrator, Denoiser, SAFEXMLAnnotation  # noqa: E402

rasterio = pytest.importorskip("rasterio")


dirname_suffix = "20190201T024655_20190201T024720_025730_02DC2A_AE07"
filename_suffix = "20190201t024655-20190201t024720-025730-02dc2a"

START_TIME = datetime(2019, 2, 1, 2, 46, 55)
END_TIME = datetime(2019, 2, 1, 2, 47, 20)

@pytest.fixture(scope="module")
def granule_directory(tmp_path_factory):
  """Create a granule directory."""
  data_dir = tmp_path_factory.mktemp("data")
  gdir = data_dir / f"S1A_IW_GRDH_1SDV_{dirname_suffix}.SAFE"
  os.mkdir(gdir)
  return gdir


@pytest.fixture(scope="module")
def annotation_file(granule_directory):
  """Create an annotation file."""
  ann_dir = granule_directory / "annotation"
  os.makedirs(ann_dir, exist_ok=True)
  annotation_file = ann_dir / f"s1a-iw-grd-vv-{filename_suffix}-001.xml"
  with open(annotation_file, "wb") as fd:
      fd.write(annotation_xml)
  return annotation_file


@pytest.fixture(scope="module")
def annotation_filehandler(annotation_file):
  """Create an annotation filehandler."""
  filename_info = dict(start_time=START_TIME, end_time=END_TIME, polarization="vv")
  return SAFEXMLAnnotation(annotation_file, filename_info, None)


@pytest.fixture(scope="module")
def calibration_file(granule_directory):
  """Create a calibration file."""
  cal_dir = granule_directory / "annotation" / "calibration"
  os.makedirs(cal_dir, exist_ok=True)
  calibration_file = cal_dir / f"calibration-s1a-iw-grd-vv-{filename_suffix}-001.xml"
  with open(calibration_file, "wb") as fd:
      fd.write(calibration_xml)
  return Path(calibration_file)

@pytest.fixture(scope="module")
def calibration_filehandler(calibration_file, annotation_filehandler):
  """Create a calibration filehandler."""
  filename_info = dict(start_time=START_TIME, end_time=END_TIME, polarization="vv")
  return Calibrator(calibration_file,
                    filename_info,
                    None,
                    image_shape=annotation_filehandler.image_shape)

@pytest.fixture(scope="module")
def noise_file(granule_directory):
  """Create a noise file."""
  noise_dir = granule_directory / "annotation" / "calibration"
  os.makedirs(noise_dir, exist_ok=True)
  noise_file = noise_dir / f"noise-s1a-iw-grd-vv-{filename_suffix}-001.xml"
  with open(noise_file, "wb") as fd:
      fd.write(noise_xml)
  return noise_file


@pytest.fixture(scope="module")
def noise_filehandler(noise_file, annotation_filehandler):
  """Create a noise filehandler."""
  filename_info = dict(start_time=START_TIME, end_time=END_TIME, polarization="vv")
  return Denoiser(noise_file, filename_info, None, image_shape=annotation_filehandler.image_shape)


@pytest.fixture(scope="module")
def noise_with_holes_filehandler(annotation_filehandler, tmpdir_factory):
  """Create a noise filehandler from data with holes."""
  filename_info = dict(start_time=START_TIME, end_time=END_TIME, polarization="vv")
  noise_xml_file = tmpdir_factory.mktemp("data").join("noise_with_holes.xml")
  with open(noise_xml_file, "wb") as fd:
    fd.write(noise_xml_with_holes)
  noise_filehandler = Denoiser(noise_xml_file,
                               filename_info, None,
                               image_shape=annotation_filehandler.image_shape)
  return noise_filehandler



@pytest.fixture(scope="module")
def measurement_file(granule_directory):
  """Create a tiff measurement file."""
  GCP = rasterio.control.GroundControlPoint

  gcps = [GCP(0, 0, 0, 0, 0),
          GCP(0, 3, 1, 0, 0),
          GCP(3, 0, 0, 1, 0),
          GCP(3, 3, 1, 1, 0),
          GCP(0, 7, 2, 0, 0),
          GCP(3, 7, 2, 1, 0),
          GCP(7, 7, 2, 2, 0),
          GCP(7, 3, 1, 2, 0),
          GCP(7, 0, 0, 2, 0),
          GCP(0, 15, 3, 0, 0),
          GCP(3, 15, 3, 1, 0),
          GCP(7, 15, 3, 2, 0),
          GCP(15, 15, 3, 3, 0),
          GCP(15, 7, 2, 3, 0),
          GCP(15, 3, 1, 3, 0),
          GCP(15, 0, 0, 3, 0),
          ]
  Z = np.linspace(0, 30000, 100, dtype=np.uint16).reshape((10, 10))
  m_dir = granule_directory / "measurement"
  os.makedirs(m_dir, exist_ok=True)
  filename = m_dir / f"s1a-iw-grd-vv-{filename_suffix}-001.tiff"
  with rasterio.open(
    filename,
    "w",
    driver="GTiff",
    height=Z.shape[0],
    width=Z.shape[1],
    count=1,
    dtype=Z.dtype,
    crs="+proj=latlong",
    gcps=gcps) as dst:
      dst.write(Z, 1)
  return Path(filename)


@pytest.fixture(scope="module")
def measurement_filehandler(measurement_file, noise_filehandler, calibration_filehandler):
  """Create a measurement filehandler."""
  filename_info = {"mission_id": "S1A", "dataset_name": "foo", "start_time": START_TIME, "end_time": END_TIME,
                   "polarization": "vv"}
  filetype_info = None
  from satpy.readers.sar_c_safe import SAFEGRD
  filehandler =  SAFEGRD(measurement_file,
                         filename_info,
                         filetype_info,
                         calibration_filehandler,
                         noise_filehandler)
  return filehandler



expected_longitudes = np.array([[-0., 0.54230055, 0.87563228, 1., 0.91541479,
                                 0.62184442, 0.26733714, -0., -0.18015287, -0.27312165],
                                [1.0883956 , 1.25662247, 1.34380634, 1.34995884, 1.2750712 ,
                                 1.11911385, 0.9390845 , 0.79202785, 0.67796547, 0.59691204],
                                [1.75505196, 1.74123364, 1.71731849, 1.68330292, 1.63918145,
                                 1.58494674, 1.52376394, 1.45880655, 1.39007883, 1.31758574],
                                [2., 1.99615628, 1.99615609, 2., 2.00768917,
                                 2.0192253 , 2.02115051, 2.        , 1.95576762, 1.88845002],
                                [1.82332931, 2.02143515, 2.18032829, 2.30002491, 2.38053511,
                                 2.4218612 , 2.43113105, 2.41546985, 2.37487052, 2.3093278 ],
                                [1.22479001, 1.81701462, 2.26984318, 2.58335874, 2.75765719,
                                 2.79279164, 2.75366973, 2.70519769, 2.64737395, 2.58019762],
                                [0.51375081, 1.53781389, 2.3082042 , 2.82500549, 3.0885147 ,
                                 3.09893859, 2.98922885, 2.89232293, 2.8082302 , 2.7369586 ],
                                [0., 1.33889733, 2.33891557, 3., 3.32266837,
                                 3.30731797, 3.1383157 , 3., 2.8923933 , 2.81551297],
                                [-0.31638932, 1.22031759, 2.36197571, 3.10836734, 3.46019271,
                                 3.41800603, 3.20098223, 3.02826595, 2.89989242, 2.81588745],
                                [-0.43541441, 1.18211505, 2.37738272, 3.1501186 , 3.50112948,
                                 3.43104055, 3.17724665, 2.97712796, 2.83072911, 2.73808164]])


class Calibration(Enum):
    """Calibration levels."""

    gamma = 1
    sigma_nought = 2
    beta_nought = 3
    dn = 4


class TestSAFEGRD:
    """Test the SAFE GRD file handler."""

    def test_read_calibrated_natural(self, measurement_filehandler):
        """Test the calibration routines."""
        calibration = Calibration.sigma_nought
        xarr = measurement_filehandler.get_dataset(DataQuery(name="measurement", polarization="vv",
                                                   calibration=calibration, quantity="natural"), info=dict())
        expected = np.array([[np.nan, 0.02707529], [2.55858416, 3.27611055]], dtype=np.float32)
        np.testing.assert_allclose(xarr.values[:2, :2], expected, rtol=2e-7)
        assert xarr.dtype == np.float32
        assert xarr.compute().dtype == np.float32

    def test_read_calibrated_dB(self, measurement_filehandler):
        """Test the calibration routines."""
        calibration = Calibration.sigma_nought
        xarr = measurement_filehandler.get_dataset(DataQuery(name="measurement", polarization="vv",
                                                   calibration=calibration, quantity="dB"), info=dict())
        expected = np.array([[np.nan, -15.674268], [4.079997, 5.153585]], dtype=np.float32)
        np.testing.assert_allclose(xarr.values[:2, :2], expected, rtol=1e-6)
        assert xarr.dtype == np.float32
        assert xarr.compute().dtype == np.float32

    def test_read_lon_lats(self, measurement_filehandler):
        """Test reading lons and lats."""
        query = DataQuery(name="longitude", polarization="vv")
        xarr = measurement_filehandler.get_dataset(query, info=dict())
        np.testing.assert_allclose(xarr.values, expected_longitudes)
        assert xarr.dtype == np.float64
        assert xarr.compute().dtype == np.float64


annotation_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<product>
  <adsHeader>
    <missionId>S1B</missionId>
    <productType>GRD</productType>
    <polarisation>HH</polarisation>
    <mode>EW</mode>
    <swath>EW</swath>
    <startTime>2020-03-15T05:04:28.137817</startTime>
    <stopTime>2020-03-15T05:05:32.416171</stopTime>
    <absoluteOrbitNumber>20698</absoluteOrbitNumber>
    <missionDataTakeId>160707</missionDataTakeId>
    <imageNumber>001</imageNumber>
  </adsHeader>
  <imageAnnotation>
    <imageInformation>
      <productFirstLineUtcTime>2020-03-15T05:04:28.137817</productFirstLineUtcTime>
      <productLastLineUtcTime>2020-03-15T05:05:32.416171</productLastLineUtcTime>
      <ascendingNodeTime>2020-03-15T04:33:22.256260</ascendingNodeTime>
      <anchorTime>2020-03-15T05:04:28.320641</anchorTime>
      <productComposition>Slice</productComposition>
      <sliceNumber>1</sliceNumber>
      <sliceList count="3">
        <slice>
          <sliceNumber>1</sliceNumber>
          <sensingStartTime>2020-03-15T05:04:29.485847</sensingStartTime>
          <sensingStopTime>2020-03-15T05:05:36.317420</sensingStopTime>
        </slice>
        <slice>
          <sliceNumber>2</sliceNumber>
          <sensingStartTime>2020-03-15T05:05:30.253413</sensingStartTime>
          <sensingStopTime>2020-03-15T05:06:34.046608</sensingStopTime>
        </slice>
        <slice>
          <sliceNumber>3</sliceNumber>
          <sensingStartTime>2020-03-15T05:06:31.020979</sensingStartTime>
          <sensingStopTime>2020-03-15T05:07:31.775796</sensingStopTime>
        </slice>
      </sliceList>
      <slantRangeTime>4.955163637998161e-03</slantRangeTime>
      <pixelValue>Detected</pixelValue>
      <outputPixels>16 bit Unsigned Integer</outputPixels>
      <rangePixelSpacing>4.000000e+01</rangePixelSpacing>
      <azimuthPixelSpacing>4.000000e+01</azimuthPixelSpacing>
      <azimuthTimeInterval>5.998353361537205e-03</azimuthTimeInterval>
      <azimuthFrequency>3.425601970000000e+02</azimuthFrequency>
      <numberOfSamples>10</numberOfSamples>
      <numberOfLines>10</numberOfLines>
      <zeroDopMinusAcqTime>-1.366569000000000e+00</zeroDopMinusAcqTime>
      <incidenceAngleMidSwath>3.468272707039038e+01</incidenceAngleMidSwath>
      <imageStatistics>
        <outputDataMean>
          <re>4.873919e+02</re>
          <im>0.000000e+00</im>
        </outputDataMean>
        <outputDataStdDev>
          <re>2.451083e+02</re>
          <im>0.000000e+00</im>
        </outputDataStdDev>
      </imageStatistics>
    </imageInformation>
  </imageAnnotation>
  <geolocationGrid>
    <geolocationGridPointList count="4">
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>0</line>
        <pixel>0</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>0</line>
        <pixel>9</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>9</line>
        <pixel>0</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>9</line>
        <pixel>9</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
    </geolocationGridPointList>
  </geolocationGrid>
</product>
"""

noise_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<noise>
  <noiseRangeVectorList count="3">
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>0</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 6.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>5</line>
      <pixel count="6">0 2 4 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>9</line>
      <pixel count="6">0 2 5 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 5.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
  </noiseRangeVectorList>
  <noiseAzimuthVectorList count="8">
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>0</firstAzimuthLine>
      <firstRangeSample>1</firstRangeSample>
      <lastAzimuthLine>1</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="1">0</line>
      <noiseAzimuthLut count="1">1.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>1</lastRangeSample>
      <line count="4">2 4 6 8</line>
      <noiseAzimuthLut count="4">2.000000e+00 2.000000e+00 2.000000e+00 2.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>4</lastRangeSample>
      <line count="2">2 4</line>
      <noiseAzimuthLut count="2">3.000000e+00 3.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>5</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>8</lastRangeSample>
      <line count="2">2 4</line>
      <noiseAzimuthLut count="2">4.000000e+00 4.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="2">5 6</line>
      <noiseAzimuthLut count="2">5.000000e+00 5.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>6</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="2">5 6</line>
      <noiseAzimuthLut count="2">6.000000e+00 6.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>6</lastRangeSample>
      <line count="1">8</line>
      <noiseAzimuthLut count="1">7.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>7</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="1">8</line>
      <noiseAzimuthLut count="1">8.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
  </noiseAzimuthVectorList>
</noise>
"""

noise_xml_with_holes = b"""<?xml version="1.0" encoding="UTF-8"?>
<noise>
  <noiseRangeVectorList count="3">
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>0</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 6.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>5</line>
      <pixel count="6">0 2 4 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>9</line>
      <pixel count="6">0 2 5 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 5.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
  </noiseRangeVectorList>
  <noiseAzimuthVectorList count="12">
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>0</firstAzimuthLine>
      <firstRangeSample>3</firstRangeSample>
      <lastAzimuthLine>2</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="1">0</line>
      <noiseAzimuthLut count="1">1.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>1</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>5</lastAzimuthLine>
      <lastRangeSample>1</lastRangeSample>
      <line count="4">2 4 5</line>
      <noiseAzimuthLut count="4">2.000000e+00 2.000000e+00 2.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>8</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="2">2 4</line>
      <noiseAzimuthLut count="2">3.000000e+00 3.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>3</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>5</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="2">3 5</line>
      <noiseAzimuthLut count="2">4.000000e+00 4.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>3</firstAzimuthLine>
      <firstRangeSample>4</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="2">3 4</line>
      <noiseAzimuthLut count="2">5.000000e+00 5.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>4</firstAzimuthLine>
      <firstRangeSample>6</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>7</lastRangeSample>
      <line count="2">4</line>
      <noiseAzimuthLut count="2">6.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>4</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>6</lastRangeSample>
      <line count="1">5 7</line>
      <noiseAzimuthLut count="1">7.000000e+00 7.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>7</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="1">6</line>
      <noiseAzimuthLut count="1">8.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>6</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="2">6 7</line>
      <noiseAzimuthLut count="2">9.000000e+00 9.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>0</lastRangeSample>
      <line count="2">8</line>
      <noiseAzimuthLut count="2">10.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="1">8 9</line>
      <noiseAzimuthLut count="1">11.000000e+00 11.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>4</firstRangeSample>
      <lastAzimuthLine>8</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="1">8</line>
      <noiseAzimuthLut count="1">12.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
  </noiseAzimuthVectorList>
</noise>
"""


calibration_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<calibration>
  <adsHeader>
    <missionId>S1A</missionId>
    <productType>GRD</productType>
    <polarisation>VV</polarisation>
    <mode>IW</mode>
    <swath>IW</swath>
    <startTime>2018-02-12T03:24:58.493726</startTime>
    <stopTime>2018-02-12T03:25:01.493726</stopTime>
    <absoluteOrbitNumber>20568</absoluteOrbitNumber>
    <missionDataTakeId>144162</missionDataTakeId>
    <imageNumber>001</imageNumber>
  </adsHeader>
  <calibrationInformation>
    <absoluteCalibrationConstant>1.000000e+00</absoluteCalibrationConstant>
  </calibrationInformation>
  <calibrationVectorList count="4">
    <calibrationVector>
      <azimuthTime>2018-02-12T03:24:58.493726</azimuthTime>
      <line>0</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
    <calibrationVector>
      <azimuthTime>2018-02-12T03:24:59.493726</azimuthTime>
      <line>3</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
    <calibrationVector>
      <azimuthTime>2018-02-12T03:25:00.493726</azimuthTime>
      <line>6</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
    <calibrationVector>
      <azimuthTime>2018-02-12T03:25:01.493726</azimuthTime>
      <line>9</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
  </calibrationVectorList>
</calibration>
"""


class TestSAFEXMLNoise:
    """Test the SAFE XML Noise file handler."""

    def setup_method(self):
        """Set up the test case."""
        self.expected_azimuth_noise = np.array([[np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                                                [np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                                                [2, 2, 3, 3, 3, 4, 4, 4, 4, np.nan],
                                                [2, 2, 3, 3, 3, 4, 4, 4, 4, np.nan],
                                                [2, 2, 3, 3, 3, 4, 4, 4, 4, np.nan],
                                                [2, 2, 5, 5, 5, 5, 6, 6, 6, 6],
                                                [2, 2, 5, 5, 5, 5, 6, 6, 6, 6],
                                                [2, 2, 5, 5, 5, 5, 6, 6, 6, 6],
                                                [2, 2, 7, 7, 7, 7, 7, 8, 8, 8],
                                                [2, 2, 7, 7, 7, 7, 7, 8, 8, 8],
                                                ])

        self.expected_range_noise = np.array([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              ])

        self.expected_azimuth_noise_with_holes = np.array(
            [[np.nan, np.nan, np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan],
             [2, 2, np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan],
             [2, 2, np.nan, 1, 1, 1, np.nan, np.nan, 3, 3],
             [2, 2, 4, 4, 5, 5, np.nan, np.nan, 3, 3],
             [2, 2, 4, 4, 5, 5, 6, 6, 3, 3],
             [2, 2, 4, 4, 7, 7, 7, 8, 8, 8],
             [9, 9, 9, 9, 7, 7, 7, 8, 8, 8],
             [9, 9, 9, 9, 7, 7, 7, 8, 8, 8],
             [10, np.nan, 11, 11, 12, 12, np.nan, np.nan, np.nan, np.nan],
             [10, np.nan, 11, 11, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
             ])

    def test_azimuth_noise_array(self, noise_filehandler):
        """Test reading the azimuth-noise array."""
        res = noise_filehandler.azimuth_noise_reader.read_azimuth_noise_array()
        np.testing.assert_array_equal(res, self.expected_azimuth_noise)

    def test_azimuth_noise_array_with_holes(self, noise_with_holes_filehandler):
        """Test reading the azimuth-noise array."""
        res = noise_with_holes_filehandler.azimuth_noise_reader.read_azimuth_noise_array()
        np.testing.assert_array_equal(res, self.expected_azimuth_noise_with_holes)

    def test_range_noise_array(self, noise_filehandler):
        """Test reading the range-noise array."""
        res = noise_filehandler.read_range_noise_array(chunks=5)
        np.testing.assert_allclose(res, self.expected_range_noise)

    def test_get_noise_dataset(self, noise_filehandler):
        """Test using get_dataset for the noise."""
        query = DataQuery(name="noise", polarization="vv")
        res = noise_filehandler.get_dataset(query, {})
        np.testing.assert_allclose(res, self.expected_azimuth_noise * self.expected_range_noise)
        assert res.dtype == np.float32
        assert res.compute().dtype == np.float32

    def test_get_noise_dataset_has_right_chunk_size(self, noise_filehandler):
        """Test using get_dataset for the noise has right chunk size in result."""
        query = DataQuery(name="noise", polarization="vv")
        res = noise_filehandler.get_dataset(query, {}, chunks=3)
        assert res.data.chunksize == (3, 3)


class TestSAFEXMLCalibration:
    """Test the SAFE XML Calibration file handler."""

    def setup_method(self):
      """Set up testing."""
      self.expected_gamma = np.array([[1840.695, 1779.672, 1718.649, 1452.926, 1187.203, 1186.226,
                                1185.249, 1184.276, 1183.303, 1181.365]]) * np.ones((10, 1))


    def test_dn_calibration_array(self, calibration_filehandler):
        """Test reading the dn calibration array."""
        expected_dn = np.ones((10, 10)) * 1087
        res = calibration_filehandler.get_calibration(Calibration.dn, chunks=5)
        np.testing.assert_allclose(res, expected_dn)
        assert res.dtype == np.float32
        assert res.compute().dtype == np.float32

    def test_beta_calibration_array(self, calibration_filehandler):
        """Test reading the beta calibration array."""
        expected_beta = np.ones((10, 10)) * 1087
        res = calibration_filehandler.get_calibration(Calibration.beta_nought, chunks=5)
        np.testing.assert_allclose(res, expected_beta)
        assert res.dtype == np.float32
        assert res.compute().dtype == np.float32

    def test_sigma_calibration_array(self, calibration_filehandler):
        """Test reading the sigma calibration array."""
        expected_sigma = np.array([[1894.274, 1841.4335, 1788.593, 1554.4165, 1320.24, 1299.104,
                                    1277.968, 1277.968, 1277.968, 1277.968]]) * np.ones((10, 1))
        res = calibration_filehandler.get_calibration(Calibration.sigma_nought, chunks=5)
        np.testing.assert_allclose(res, expected_sigma)
        assert res.dtype == np.float32
        assert res.compute().dtype == np.float32

    def test_gamma_calibration_array(self, calibration_filehandler):
        """Test reading the gamma calibration array."""
        res = calibration_filehandler.get_calibration(Calibration.gamma, chunks=5)
        np.testing.assert_allclose(res, self.expected_gamma)
        assert res.dtype == np.float32
        assert res.compute().dtype == np.float32

    def test_get_calibration_dataset(self, calibration_filehandler):
        """Test using get_dataset for the calibration."""
        query = DataQuery(name="gamma", polarization="vv")
        res = calibration_filehandler.get_dataset(query, {})
        np.testing.assert_allclose(res, self.expected_gamma)
        assert res.dtype == np.float32
        assert res.compute().dtype == np.float32

    def test_get_calibration_dataset_has_right_chunk_size(self, calibration_filehandler):
        """Test using get_dataset for the calibration yields array with right chunksize."""
        query = DataQuery(name="gamma", polarization="vv")
        res = calibration_filehandler.get_dataset(query, {}, chunks=3)
        assert res.data.chunksize == (3, 3)
        np.testing.assert_allclose(res, self.expected_gamma)

    def test_get_calibration_constant(self, calibration_filehandler):
        """Test getting the calibration constant."""
        query = DataQuery(name="calibration_constant", polarization="vv")
        res = calibration_filehandler.get_dataset(query, {})
        assert res == 1
        assert type(res) is np.float32


def test_incidence_angle(annotation_filehandler):
  """Test reading the incidence angle in an annotation file."""
  query = DataQuery(name="incidence_angle", polarization="vv")
  res = annotation_filehandler.get_dataset(query, {})
  np.testing.assert_allclose(res, 19.18318046)
  assert res.dtype == np.float32
  assert res.compute().dtype == np.float32


def test_reading_from_reader(measurement_file, calibration_file, noise_file, annotation_file):
  """Test reading using the reader defined in the config."""
  with open(Path(PACKAGE_CONFIG_PATH) / "readers" / "sar-c_safe.yaml") as fd:
    config = yaml.load(fd, Loader=yaml.UnsafeLoader)
  reader_class = config["reader"]["reader"]
  reader = reader_class(config)

  files = [measurement_file, calibration_file, noise_file, annotation_file]
  reader.create_storage_items(files)
  query = DataQuery(name="measurement", polarization="vv",
                    calibration="sigma_nought", quantity="dB")
  query = DataID(reader._id_keys, **query.to_dict())
  dataset_dict = reader.load([query])
  array = dataset_dict["measurement"]
  np.testing.assert_allclose(array.attrs["area"].lons, expected_longitudes)
  expected_db = np.array([[np.nan, -15.674268], [4.079997, 5.153585]])
  np.testing.assert_allclose(array.values[:2, :2], expected_db, rtol=1e-6)
  assert array.dtype == np.float32
  assert array.compute().dtype == np.float32


def test_filename_filtering_from_reader(measurement_file, calibration_file, noise_file, annotation_file, tmp_path):
  """Test that filenames get filtered before filehandlers are created."""
  with open(Path(PACKAGE_CONFIG_PATH) / "readers" / "sar-c_safe.yaml") as fd:
    config = yaml.load(fd, Loader=yaml.UnsafeLoader)
  reader_class = config["reader"]["reader"]
  filter_parameters = {"start_time": datetime(2019, 2, 1, 0, 0, 0),
                       "end_time": datetime(2019, 2, 1, 12, 0, 0)}
  reader = reader_class(config, filter_parameters)

  spurious_file = (tmp_path / "S1A_IW_GRDH_1SDV_20190202T024655_20190202T024720_025730_02DC2A_AE07.SAFE" /
                   "measurement" /
                   "s1a-iw-grd-vv-20190202t024655-20190202t024720-025730-02dc2a-001.tiff")


  files = [spurious_file, measurement_file, calibration_file, noise_file, annotation_file]

  files = reader.filter_selected_filenames(files)
  assert spurious_file not in files
  try:
    reader.create_storage_items(files)
  except rasterio.RasterioIOError as err:
     pytest.fail(str(err))


def test_swath_def_contains_gcps_and_bounding_box(measurement_file, calibration_file, noise_file, annotation_file):
  """Test reading using the reader defined in the config."""
  with open(Path(PACKAGE_CONFIG_PATH) / "readers" / "sar-c_safe.yaml") as fd:
    config = yaml.load(fd, Loader=yaml.UnsafeLoader)
  reader_class = config["reader"]["reader"]
  reader = reader_class(config)

  files = [measurement_file, calibration_file, noise_file, annotation_file]
  reader.create_storage_items(files)
  query = DataQuery(name="measurement", polarization="vv",
                    calibration="sigma_nought", quantity="dB")
  query = DataID(reader._id_keys, **query.to_dict())
  dataset_dict = reader.load([query])
  array = dataset_dict["measurement"]
  assert array.attrs["area"].attrs["gcps"] is not None
  assert array.attrs["area"].attrs["bounding_box"] is not None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""SAFE MSI L1C/L2A reader.

The MSI data has a special value for saturated pixels. By default, these
pixels are set to np.inf, but for some applications it might be desirable
to have these pixels left untouched.
For this case, the `mask_saturated` flag is available in the reader, and can be
toggled with ``reader_kwargs`` upon Scene creation::

    scene = satpy.Scene(filenames,
                        reader='msi_safe',
                        reader_kwargs={'mask_saturated': False})
    scene.load(['B01'])

L1C/L2A format description for the files read here:

  https://sentinels.copernicus.eu/documents/247904/685211/S2-PDGS-TAS-DI-PSD-V14.9.pdf/3d3b6c9c-4334-dcc4-3aa7-f7c0deffbaf7?t=1643013091529

NOTE: At present, L1B data is not supported. If the user needs radiance data instead of counts or reflectances, these
are retrieved by first calculating the reflectance and then working back to the radiance. L1B radiance data support
will be added once the data is published onto the Copernicus data ecosystem.

"""

import logging
from datetime import datetime

import dask.array as da
import defusedxml.ElementTree as ET
import numpy as np
import xarray as xr
from pyresample import geometry

from satpy._compat import cached_property
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()

PLATFORMS = {"S2A": "Sentinel-2A",
             "S2B": "Sentinel-2B",
             "S2C": "Sentinel-2C",
             "S2D": "Sentinel-2D"}


class SAFEMSIL1C(BaseFileHandler):
    """File handler for SAFE MSI files (jp2)."""

    def __init__(self, filename, filename_info, filetype_info, mda, tile_mda,
                 mask_saturated=True):
        """Initialize the reader."""
        super(SAFEMSIL1C, self).__init__(filename, filename_info,
                                         filetype_info)
        del mask_saturated
        self._channel = filename_info["band_name"]
        self.process_level = filename_info["process_level"]
        if self.process_level not in ["L1C", "L2A"]:
            raise ValueError(f"Unsupported process level: {self.process_level}")
        self._tile_mda = tile_mda
        self._mda = mda
        self.platform_name = PLATFORMS[filename_info["fmission_id"]]
        self._start_time = self._tile_mda.start_time()
        self._end_time = filename_info["observation_time"]

    def get_dataset(self, key, info):
        """Load a dataset."""
        if self._channel != key["name"]:
            return

        logger.debug("Reading %s.", key["name"])

        proj = self._read_from_file(key)
        if proj is None:
            return
        proj.attrs = info.copy()
        proj.attrs["platform_name"] = self.platform_name
        return proj

    def _read_from_file(self, key):
        proj = xr.open_dataset(self.filename, engine="rasterio", chunks=CHUNK_SIZE)["band_data"]
        proj = proj.squeeze("band")
        if key["calibration"] == "reflectance":
            return self._mda.calibrate_to_reflectances(proj, self._channel)
        if key["calibration"] == "radiance":
            # The calibration procedure differs for L1B and L1C/L2A data!
            if self.process_level in ["L1C", "L2A"]:
                # For higher level data, radiances must be computed from the reflectance.
                # By default, we use the mean solar angles so that the user does not need to resample,
                # but the user can also choose to use the solar angles from the tile metadata.
                # This is on a coarse grid so for most bands must be resampled before use.
                dq = dict(name="solar_zenith_angle", resolution=key["resolution"])
                zen = self._tile_mda.get_dataset(dq, dict(xml_tag="Sun_Angles_Grid/Zenith"))
                tmp_refl = self._mda.calibrate_to_reflectances(proj, self._channel)
                return self._mda.calibrate_to_radiances(tmp_refl, zen, self._channel)
            else:
                # For L1B the radiances can be directly computed from the digital counts.
                return self._mda.calibrate_to_radiances_l1b(proj, self._channel)


        if key["calibration"] == "counts":
            return self._mda._sanitize_data(proj)
        if key["calibration"] in ["aerosol_thickness", "water_vapor"]:
            return self._mda.calibrate_to_atmospheric(proj, self._channel)

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._start_time

    def get_area_def(self, dsid):
        """Get the area def."""
        if self._channel != dsid["name"]:
            return

        return self._tile_mda.get_area_def(dsid)


class SAFEMSIXMLMetadata(BaseFileHandler):
    """Base class for SAFE MSI XML metadata filehandlers."""

    def __init__(self, filename, filename_info, filetype_info, mask_saturated=True):
        """Init the reader."""
        super().__init__(filename, filename_info, filetype_info)
        self._start_time = filename_info["observation_time"]
        self._end_time = filename_info["observation_time"]
        self.root = ET.parse(self.filename)
        self.tile = filename_info["dtile_number"]
        self.process_level = filename_info["process_level"]
        self.platform_name = PLATFORMS[filename_info["fmission_id"]]
        self.mask_saturated = mask_saturated
        import bottleneck  # noqa
        import geotiepoints  # noqa

    @property
    def end_time(self):
        """Get end time."""
        return self._start_time

    @property
    def start_time(self):
        """Get start time."""
        return self._start_time


class SAFEMSIMDXML(SAFEMSIXMLMetadata):
    """File handle for sentinel 2 safe XML generic metadata."""

    def calibrate_to_reflectances(self, data, band_name):
        """Calibrate *data* using the radiometric information for the metadata."""
        quantification = int(self.root.find(".//QUANTIFICATION_VALUE").text) if self.process_level[:2] == "L1" else \
            int(self.root.find(".//BOA_QUANTIFICATION_VALUE").text)
        data = self._sanitize_data(data)
        return (data + self.band_offset(band_name)) / quantification * 100

    def calibrate_to_atmospheric(self, data, band_name):
        """Calibrate L2A AOT/WVP product."""
        atmospheric_bands = ["AOT", "WVP"]
        if self.process_level == "L1C" or self.process_level == "L1B":
            return
        elif self.process_level == "L2A" and band_name not in atmospheric_bands:
            return

        quantification = float(self.root.find(f".//{band_name}_QUANTIFICATION_VALUE").text)
        data = self._sanitize_data(data)
        return data / quantification

    def _sanitize_data(self, data):
        data = data.where(data != self.no_data)
        if self.mask_saturated:
            data = data.where(data != self.saturated, np.inf)
        return data

    def band_offset(self, band):
        """Get the band offset for *band*."""
        band_index = self._band_index(band)
        return self.band_offsets.get(band_index, 0)

    def _band_index(self, band):
        band_indices = self.band_indices
        band_conversions = {"B01": "B1", "B02": "B2", "B03": "B3", "B04": "B4", "B05": "B5", "B06": "B6", "B07": "B7",
                            "B08": "B8", "B8A": "B8A", "B09": "B9", "B10": "B10", "B11": "B11", "B12": "B12"}
        band_index = band_indices[band_conversions[band]]
        return band_index

    @cached_property
    def band_indices(self):
        """Get the band indices from the metadata."""
        spectral_info = self.root.findall(".//Spectral_Information")
        band_indices = {spec.attrib["physicalBand"]: int(spec.attrib["bandId"]) for spec in spectral_info}
        return band_indices

    @cached_property
    def band_offsets(self):
        """Get the band offsets from the metadata."""
        offsets = self.root.find(".//Radiometric_Offset_List") if self.process_level[:2] == "L1" else \
            self.root.find(".//BOA_ADD_OFFSET_VALUES_LIST")
        if offsets is not None:
            band_offsets = {int(off.attrib["band_id"]): float(off.text) for off in offsets}
        else:
            band_offsets = {}
        return band_offsets

    def solar_irradiance(self, band_name):
        """Get the solar irradiance for a given *band_name*."""
        band_index = self._band_index(band_name)
        return self.solar_irradiances[band_index]

    @cached_property
    def solar_irradiances(self):
        """Get the TOA solar irradiance values from the metadata."""
        irrads = self.root.find(".//Solar_Irradiance_List")

        if irrads is not None:
            solar_irrad = {int(irr.attrib["bandId"]): float(irr.text) for irr in irrads}
        if len(solar_irrad) > 0:
            return solar_irrad
        raise ValueError("No solar irradiance values were found in the metadata.")

    @cached_property
    def solar_correction_factor(self):
        """Get the solar correction factor from the metadata."""
        sed = self.root.find(".//U")
        if sed.text is not None:
            return float(sed.text)
        raise ValueError("Solar correction factor, U, in metadata is missing.")

    @cached_property
    def special_values(self):
        """Get the special values from the metadata."""
        special_values = self.root.findall(".//Special_Values")
        special_values_dict = {value[0].text: float(value[1].text) for value in special_values}
        return special_values_dict

    @property
    def no_data(self):
        """Get the nodata value from the metadata."""
        return self.special_values["NODATA"]

    @property
    def saturated(self):
        """Get the saturated value from the metadata."""
        return self.special_values["SATURATED"]

    def calibrate_to_radiances_l1b(self, data, band_name):
        """Calibrate *data* to radiance using the radiometric information for the metadata."""
        physical_gain = self.physical_gain(band_name)
        data = self._sanitize_data(data)
        return (data + self.band_offset(band_name)) / physical_gain

    def calibrate_to_radiances(self, data, solar_zenith, band_name):
        """Calibrate *data* to radiance using the radiometric information for the metadata.

        This follows the principle set out in the user guide:
        https://sentiwiki.copernicus.eu/web/s2-products#S2Products-Level-1CProductsS2-Products-L1Ctrue
        """
        ucor = self.solar_correction_factor
        solar_irrad_band = self.solar_irradiance(band_name)

        solar_zenith = np.deg2rad(solar_zenith)

        return (data / 100.) * solar_irrad_band * np.cos(solar_zenith) * ucor / np.pi

    def physical_gain(self, band_name):
        """Get the physical gain for a given *band_name*."""
        band_index = self._band_index(band_name)
        return self.physical_gains[band_index]

    @cached_property
    def physical_gains(self):
        """Get the physical gains dictionary."""
        physical_gains = {int(elt.attrib["bandId"]): float(elt.text) for elt in self.root.findall(".//PHYSICAL_GAINS")}
        return physical_gains


def _fill_swath_edges(angles):
    """Fill gaps at edges of swath."""
    darr = xr.DataArray(angles, dims=["y", "x"])
    darr = darr.bfill("x")
    darr = darr.ffill("x")
    darr = darr.bfill("y")
    darr = darr.ffill("y")
    angles = darr.data
    return angles


class SAFEMSITileMDXML(SAFEMSIXMLMetadata):
    """File handle for sentinel 2 safe XML tile metadata."""

    def __init__(self, filename, filename_info, filetype_info, mask_saturated=True):
        """Init the reader."""
        super().__init__(filename, filename_info, filetype_info, mask_saturated)
        self.geocoding = self.root.find(".//Tile_Geocoding")

    def get_area_def(self, dsid):
        """Get the area definition of the dataset."""
        area_extent = self._area_extent(dsid["resolution"])
        cols, rows = self._shape(dsid["resolution"])
        area = geometry.AreaDefinition(
            self.tile,
            "On-the-fly area",
            self.tile,
            self.projection,
            cols,
            rows,
            area_extent)
        return area

    @cached_property
    def projection(self):
        """Get the geographic projection."""
        from pyproj import CRS
        epsg = self.geocoding.find("HORIZONTAL_CS_CODE").text
        return CRS(epsg)

    def _area_extent(self, resolution):
        cols, rows = self._shape(resolution)
        geoposition = self.geocoding.find('Geoposition[@resolution="' + str(resolution) + '"]')
        ulx = float(geoposition.find("ULX").text)
        uly = float(geoposition.find("ULY").text)
        xdim = float(geoposition.find("XDIM").text)
        ydim = float(geoposition.find("YDIM").text)
        area_extent = (ulx, uly + rows * ydim, ulx + cols * xdim, uly)
        return area_extent

    def _shape(self, resolution):
        rows = int(self.geocoding.find('Size[@resolution="' + str(resolution) + '"]/NROWS').text)
        cols = int(self.geocoding.find('Size[@resolution="' + str(resolution) + '"]/NCOLS').text)
        return cols, rows

    def start_time(self):
        """Get the observation time from the tile metadata."""
        timestr = self.root.find(".//SENSING_TIME").text
        return datetime.strptime(timestr, "%Y-%m-%dT%H:%M:%S.%fZ")

    @staticmethod
    def _do_interp(minterp, xcoord, ycoord):
        interp_points2 = np.vstack((ycoord.ravel(), xcoord.ravel()))
        res = minterp(interp_points2)
        return res.reshape(xcoord.shape)

    def interpolate_angles(self, angles, resolution):
        """Interpolate the angles."""
        from geotiepoints.multilinear import MultilinearInterpolator

        cols, rows = self._shape(resolution)

        smin = [0, 0]
        smax = np.array(angles.shape) - 1
        orders = angles.shape
        minterp = MultilinearInterpolator(smin, smax, orders)
        minterp.set_values(da.atleast_2d(angles.ravel()))

        y = da.arange(rows, dtype=angles.dtype, chunks=CHUNK_SIZE) / (rows-1) * (angles.shape[0] - 1)
        x = da.arange(cols, dtype=angles.dtype, chunks=CHUNK_SIZE) / (cols-1) * (angles.shape[1] - 1)
        xcoord, ycoord = da.meshgrid(x, y)
        return da.map_blocks(self._do_interp, minterp, xcoord, ycoord, dtype=angles.dtype, chunks=xcoord.chunks)

    def _get_coarse_dataset(self, key, info):
        """Get the coarse dataset refered to by `key` from the XML data."""
        angles = self.root.find(".//Tile_Angles")
        if key["name"] in ["solar_zenith_angle", "solar_azimuth_angle"]:
            angles = self._get_solar_angles(angles, info)
        elif key["name"] in ["satellite_zenith_angle", "satellite_azimuth_angle"]:
            angles = self._get_satellite_angles(angles, info)
        else:
            angles = None
        return angles

    def _get_solar_angles(self, angles, info):
        angles = self._get_values_from_tag(angles, info["xml_tag"])
        return angles

    @staticmethod
    def _get_values_from_tag(xml_tree, xml_tag):
        elts = xml_tree.findall(xml_tag + "/Values_List/VALUES")
        return np.array([[val for val in elt.text.split()] for elt in elts],
                        dtype=np.float64)

    def _get_satellite_angles(self, angles, info):
        arrays = []
        elts = angles.findall(info["xml_tag"] + '[@bandId="1"]')
        for elt in elts:
            arrays.append(self._get_values_from_tag(elt, info["xml_item"]))
        angles = np.nanmean(np.dstack(arrays), axis=-1)
        return angles

    def get_dataset(self, key, info):
        """Get the dataset referred to by `key`."""
        angles = self._get_coarse_dataset(key, info)
        if angles is None:
            return None

        angles = _fill_swath_edges(angles)

        res = self.interpolate_angles(angles, key["resolution"])

        proj = xr.DataArray(res, dims=["y", "x"])
        proj.attrs = info.copy()
        proj.attrs["units"] = "degrees"
        proj.attrs["platform_name"] = self.platform_name
        return proj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.viirs_l2_jrr module.

Note: This is adapted from the test_slstr_l2.py code.
"""
from __future__ import annotations

import contextlib
import datetime as dt
import shutil
import warnings
from collections.abc import Iterator
from pathlib import Path
from typing import Iterable

import dask
import dask.array as da
import numpy as np
import numpy.typing as npt
import pytest
import xarray as xr
from pyresample import SwathDefinition
from pytest import TempPathFactory  # noqa: PT013
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy.tests.utils import RANDOM_GEN

I_COLS = 6400
I_ROWS = 32  # one scan
M_COLS = 3200
M_ROWS = 16  # one scan
START_TIME = dt.datetime(2023, 5, 30, 17, 55, 41, 0)
END_TIME = dt.datetime(2023, 5, 30, 17, 57, 5, 0)
QF1_FLAG_MEANINGS = """
\tBits are listed from the MSB (bit 7) to the LSB (bit 0):
\tBit    Description
\t6-7    SUN GLINT;
\t       00 -- none
\t       01 -- geometry based
\t       10 -- wind speed based
\t       11 -- geometry & wind speed based
\t5      low sun mask;
\t       0 -- high
\t       1 -- low
\t4      day/night;
\t       0 -- day
\t       1 -- night
\t2-3    cloud detection & confidence;
\t       00 -- confident clear
\t       01 -- probably clear
\t       10 -- probably cloudy
\t       11 -- confident cloudy
\t0-1    cloud mask quality;
\t       00 -- poor
\t       01 -- low
\t       10 -- medium
\t       11 -- high
"""


@pytest.fixture(scope="module")
def surface_reflectance_file(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake surface reflectance EDR file."""
    return _create_surface_reflectance_file(tmp_path_factory, START_TIME, include_veg_indices=False)


@pytest.fixture(scope="module")
def surface_reflectance_file2(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake surface reflectance EDR file."""
    return _create_surface_reflectance_file(tmp_path_factory, START_TIME + dt.timedelta(minutes=5),
                                            include_veg_indices=False)


@pytest.fixture(scope="module")
def multiple_surface_reflectance_files(surface_reflectance_file, surface_reflectance_file2) -> list[Path]:
    """Get two multiple surface reflectance files."""
    return [surface_reflectance_file, surface_reflectance_file2]


@pytest.fixture(scope="module")
def surface_reflectance_with_veg_indices_file(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake surface reflectance EDR file with vegetation indexes included."""
    return _create_surface_reflectance_file(tmp_path_factory, START_TIME, include_veg_indices=True)


@pytest.fixture(scope="module")
def surface_reflectance_with_veg_indices_file2(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake surface reflectance EDR file with vegetation indexes included."""
    return _create_surface_reflectance_file(tmp_path_factory, START_TIME + dt.timedelta(minutes=5),
                                            include_veg_indices=True)


@pytest.fixture(scope="module")
def multiple_surface_reflectance_files_with_veg_indices(surface_reflectance_with_veg_indices_file,
                                                        surface_reflectance_with_veg_indices_file2) -> list[Path]:
    """Get two multiple surface reflectance files with vegetation indexes included."""
    return [surface_reflectance_with_veg_indices_file, surface_reflectance_with_veg_indices_file2]


def _create_surface_reflectance_file(
        tmp_path_factory: TempPathFactory,
        start_time: dt.datetime,
        include_veg_indices: bool = False,
) -> Path:
    fn = f"SurfRefl_v1r2_npp_s{start_time:%Y%m%d%H%M%S}0_e{END_TIME:%Y%m%d%H%M%S}0_c202305302025590.nc"
    sr_vars = _create_surf_refl_variables()
    if include_veg_indices:
        sr_vars.update(_create_veg_index_variables())
    return _create_fake_file(tmp_path_factory, fn, sr_vars)


def _create_surf_refl_variables() -> dict[str, xr.DataArray]:
    dim_y_750 = "Along_Track_750m"
    dim_x_750 = "Along_Scan_750m"
    m_dims = (dim_y_750, dim_x_750)
    dim_y_375 = "Along_Track_375m"
    dim_x_375 = "Along_Scan_375m"
    i_dims = (dim_y_375, dim_x_375)

    lon_attrs = {"standard_name": "longitude", "units": "degrees_east", "_FillValue": -999.9,
                 "valid_min": -180.0, "valid_max": 180.0}
    lat_attrs = {"standard_name": "latitude", "units": "degrees_north", "_FillValue": -999.9,
                 "valid_min": -90.0, "valid_max": 90.0}
    sr_attrs = {"units": "unitless", "_FillValue": -9999,
                "scale_factor": np.float32(0.0001), "add_offset": np.float32(0.0)}

    i_data = RANDOM_GEN.random((I_ROWS, I_COLS)).astype(np.float32)
    m_data = RANDOM_GEN.random((M_ROWS, M_COLS)).astype(np.float32)
    lon_i_data = (i_data * 360) - 180.0
    lon_m_data = (m_data * 360) - 180.0
    lat_i_data = (i_data * 180) - 90.0
    lat_m_data = (m_data * 180) - 90.0
    for geo_var in (lon_i_data, lon_m_data, lat_i_data, lat_m_data):
        geo_var[0, 0] = -999.9
        geo_var[0, 1] = -999.3
    data_arrs = {
        "Longitude_at_375m_resolution": xr.DataArray(lon_i_data, dims=i_dims, attrs=lon_attrs),
        "Latitude_at_375m_resolution": xr.DataArray(lat_i_data, dims=i_dims, attrs=lat_attrs),
        "Longitude_at_750m_resolution": xr.DataArray(lon_m_data, dims=m_dims, attrs=lon_attrs),
        "Latitude_at_750m_resolution": xr.DataArray(lat_m_data, dims=m_dims, attrs=lat_attrs),
        "375m Surface Reflectance Band I1": xr.DataArray(i_data, dims=i_dims, attrs=sr_attrs),
        "750m Surface Reflectance Band M1": xr.DataArray(m_data, dims=m_dims, attrs=sr_attrs),
    }
    for data_arr in data_arrs.values():
        data_arr.encoding["chunksizes"] = data_arr.shape
        if "scale_factor" not in data_arr.attrs:
            continue
        data_arr.encoding["dtype"] = np.int16
        data_arr.encoding["scale_factor"] = data_arr.attrs.pop("scale_factor")
        data_arr.encoding["add_offset"] = data_arr.attrs.pop("add_offset")
    return data_arrs


def _create_veg_index_variables() -> dict[str, xr.DataArray]:
    dim_y_750 = "Along_Track_750m"
    dim_x_750 = "Along_Scan_750m"
    m_dims = (dim_y_750, dim_x_750)
    dim_y_375 = "Along_Track_375m"
    dim_x_375 = "Along_Scan_375m"
    i_dims = (dim_y_375, dim_x_375)

    vi_data = np.zeros((I_ROWS, I_COLS), dtype=np.float32)
    vi_data[0, :7] = [-2.0, -1.0, -0.5, 0.0, 0.5, 1.0, 1.5]
    data_arrs = {
        "NDVI": xr.DataArray(vi_data, dims=i_dims, attrs={"units": "unitless"}),
        "EVI": xr.DataArray(vi_data, dims=i_dims, attrs={"units": "unitless"}),
    }
    data_arrs["NDVI"].encoding["dtype"] = np.float32
    data_arrs["EVI"].encoding["dtype"] = np.float32

    # Quality Flags are from the Surface Reflectance data, but only used for VI products in the reader
    for qf_num in range(1, 8):
        qf_name = f"QF{qf_num} Surface Reflectance"
        qf_data = np.zeros((M_ROWS, M_COLS), dtype=np.uint8)
        bad_qf_start = 4  # 0.5x the last test pixel set in "vi_data" above (I-band versus M-band index)
        if qf_num == 1:
            qf_data[:, :] |= 0b00000010  # medium cloud mask quality everywhere
            # qf_data[0, bad_qf_start] |= 0b11000000  # sun glint
            qf_data[0, bad_qf_start] |= 0b00001100  # cloudy
            qf_data[0, bad_qf_start + 1] |= 0b00001100  # cloudy
            qf_data[0, bad_qf_start + 2] = 0b00000001  # low cloud mask quality
        elif qf_num == 2:
            qf_data[:, :] |= 0b00000011  # desert everywhere
            qf_data[0, bad_qf_start + 3] |= 0b00100000  # snow or ice
            qf_data[0, bad_qf_start + 4] |= 0b00001000  # cloud shadow
            qf_data[0, bad_qf_start + 5] = 0b00000001  # deep ocean
        elif qf_num == 7:
            qf_data[0, bad_qf_start + 6] |= 0b00001100  # high aerosol
            qf_data[0, bad_qf_start + 7] |= 0b00000010  # adjacent to cloud

        data_arr = xr.DataArray(qf_data, dims=m_dims, attrs={"flag_meanings": QF1_FLAG_MEANINGS})
        data_arr.encoding["dtype"] = np.uint8
        data_arrs[qf_name] = data_arr
    return data_arrs


@pytest.fixture(scope="module")
def cloud_height_file(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake CloudHeight VIIRS EDR file."""
    fn = f"JRR-CloudHeight_v3r2_npp_s{START_TIME:%Y%m%d%H%M%S}0_e{END_TIME:%Y%m%d%H%M%S}0_c202307231023395.nc"
    data_vars = _create_continuous_variables(
        ("CldTopTemp", "CldTopHght", "CldTopPres")
    )
    lon_pc = data_vars["Longitude"].copy(deep=True)
    lat_pc = data_vars["Latitude"].copy(deep=True)
    lon_pc.attrs["long_name"] = "BAD"
    lat_pc.attrs["long_name"] = "BAD"
    del lon_pc.encoding["_FillValue"]
    del lat_pc.encoding["_FillValue"]
    data_vars["Longitude_Pc"] = lon_pc
    data_vars["Latitude_Pc"] = lat_pc
    return _create_fake_file(tmp_path_factory, fn, data_vars)


@pytest.fixture(scope="module")
def aod_file(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake AOD VIIRs EDR file."""
    fn = f"JRR-AOD_v3r2_npp_s{START_TIME:%Y%m%d%H%M%S}0_e{END_TIME:%Y%m%d%H%M%S}0_c202307231023395.nc"
    data_vars = _create_continuous_variables(
        ("AOD550",),
        data_attrs={
            "valid_range": [-0.5, 0.5],
            "units": "1",
            "_FillValue": -999.999,
        }
    )
    qc_data = np.zeros(data_vars["AOD550"].shape, dtype=np.int8)
    qc_data[-1, -1] = 2
    data_vars["QCAll"] = xr.DataArray(
        qc_data,
        dims=data_vars["AOD550"].dims,
        attrs={"valid_range": [0, 3]},
    )
    data_vars["QCAll"].encoding["_FillValue"] = -128
    return _create_fake_file(tmp_path_factory, fn, data_vars)


@pytest.fixture(scope="module")
def lst_file(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake VLST EDR file."""
    fn = f"LST_v2r0_npp_s{START_TIME:%Y%m%d%H%M%S}0_e{END_TIME:%Y%m%d%H%M%S}0_c202307241854058.nc"
    data_vars = _create_lst_variables()
    return _create_fake_file(tmp_path_factory, fn, data_vars)


def _create_lst_variables() -> dict[str, xr.DataArray]:
    data_vars = _create_continuous_variables(("VLST",))

    # VLST scale factors
    data_vars["VLST"].data = (data_vars["VLST"].data / 0.0001).astype(np.int16)
    data_vars["VLST"].encoding.pop("scale_factor")
    data_vars["VLST"].encoding.pop("add_offset")
    data_vars["LST_ScaleFact"] = xr.DataArray(np.float32(0.0001))
    data_vars["LST_Offset"] = xr.DataArray(np.float32(0.0))

    return data_vars


@pytest.fixture(scope="module")
def volcanic_ash_file(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake *partial* Volcanic Ash VIIRs EDR file."""
    fn = f"JRR-VolcanicAsh_v3r0_j01_s{START_TIME:%Y%m%d%H%M%S}0_e{END_TIME:%Y%m%d%H%M%S}0_c202307231023395.nc"
    data_vars = _create_continuous_variables(
        ("AshBeta",),
        data_attrs={
            "units": "1",
            "_FillValue": -999.,
        }
    )
    # The 'Det_QF_Size' variable is actually a scalar, but the there's no way to check it is dropped other than
    # making it 2D
    data_vars["Det_QF_Size"] = xr.DataArray(np.array([[1, 2]], dtype=np.int32),
                                            attrs={"_FillValue": -999, "units": "1"})
    return _create_fake_file(tmp_path_factory, fn, data_vars)


@pytest.fixture(scope="module")
def cloud_phase_file(tmp_path_factory: TempPathFactory) -> Path:
    """Generate fake AOD VIIRs EDR file."""
    fn = f"JRR-CloudPhase_v3r2_npp_s{START_TIME:%Y%m%d%H%M%S}0_e{END_TIME:%Y%m%d%H%M%S}0_c202307231023395.nc"

    # get lon/lat variables
    data_vars = _create_continuous_variables([])
    phase_data = (RANDOM_GEN.random((M_ROWS, M_COLS)) * 6).astype(np.int8)
    cloud_phase = xr.DataArray(
        phase_data,
        dims=("Rows", "Columns"),
        attrs={
            "valid_range": [np.int8(0), np.int8(5)],
            "units": "1",
            "coordinates": "Longitude Latitude",
        },
    )
    cloud_phase.encoding["_FillValue"] = np.int8(-128)
    cloud_phase.encoding["dtype"] = np.int8
    data_vars["CloudPhase"] = cloud_phase
    return _create_fake_file(tmp_path_factory, fn, data_vars)


def _create_continuous_variables(
        var_names: Iterable[str],
        data_attrs: None | dict = None
) -> dict[str, xr.DataArray]:
    dims = ("Rows", "Columns")

    lon_attrs = {"standard_name": "longitude", "units": "degrees_east", "_FillValue": -999.9}
    lat_attrs = {"standard_name": "latitude", "units": "degrees_north", "_FillValue": -999.9}
    cont_attrs = data_attrs
    if cont_attrs is None:
        cont_attrs = {"units": "Kelvin", "_FillValue": -9999,
                      "scale_factor": np.float32(0.0001), "add_offset": np.float32(0.0)}

    m_data = RANDOM_GEN.random((M_ROWS, M_COLS)).astype(np.float32)
    data_arrs = {
        "Longitude": xr.DataArray(m_data, dims=dims, attrs=lon_attrs),
        "Latitude": xr.DataArray(m_data, dims=dims, attrs=lat_attrs),
    }
    cont_data = m_data
    if "valid_range" in cont_attrs:
        valid_range = cont_attrs["valid_range"]
        # scale 0-1 random data to fit in valid_range
        cont_data = cont_data * (valid_range[1] - valid_range[0]) + valid_range[0]
    for var_name in var_names:
        data_arrs[var_name] = xr.DataArray(cont_data, dims=dims, attrs=cont_attrs)
    for data_arr in data_arrs.values():
        if "_FillValue" in data_arr.attrs:
            data_arr.encoding["_FillValue"] = data_arr.attrs.pop("_FillValue")
        data_arr.encoding["coordinates"] = "Longitude Latitude"
        if "scale_factor" not in data_arr.attrs:
            continue
        data_arr.encoding["dtype"] = np.int16
        data_arr.encoding["scale_factor"] = data_arr.attrs.pop("scale_factor")
        data_arr.encoding["add_offset"] = data_arr.attrs.pop("add_offset")
    return data_arrs


def _create_fake_file(tmp_path_factory: TempPathFactory, filename: str, data_arrs: dict[str, xr.DataArray]) -> Path:
    tmp_path = tmp_path_factory.mktemp("viirs_edr_tmp")
    file_path = tmp_path / filename
    ds = _create_fake_dataset(data_arrs)
    ds.to_netcdf(file_path)
    return file_path


def _create_fake_dataset(vars_dict: dict[str, xr.DataArray]) -> xr.Dataset:
    ds = xr.Dataset(
        vars_dict,
        attrs={}
    )
    return ds


def test_available_datasets(aod_file):
    """Test that available datasets doesn't claim non-filetype datasets.

    For example, if a YAML-configured dataset's file type is not loaded
    then the available status is `None` and should remain `None`. This
    means no file type knows what to do with this dataset. If it is
    `False` then that means that a file type knows of the dataset, but
    that the variable is not available in the file. In the below test
    this isn't the case so the YAML-configured dataset should be
    provided once and have a `None` availability.

    """
    from satpy.readers.viirs_edr import VIIRSJRRFileHandler
    file_handler = VIIRSJRRFileHandler(
        aod_file,
        {"platform_shortname": "npp"},
        {"file_type": "jrr_aod"},
    )
    fake_yaml_datasets = [
        (None, {"file_key": "fake", "file_type": "fake_file", "name": "fake"}),
    ]
    available_datasets = list(file_handler.available_datasets(configured_datasets=fake_yaml_datasets))
    fake_availables = [avail_tuple for avail_tuple in available_datasets if avail_tuple[1]["name"] == "fake"]
    assert len(fake_availables) == 1
    assert fake_availables[0][0] is None


@contextlib.contextmanager
def set_chunk_size(bytes_in_m_row: int) -> Iterator[None]:
    """Set dask chunks and ignore expected performance warning."""
    with dask.config.set({"array.chunk-size": f"{bytes_in_m_row * 4}B"}), \
            warnings.catch_warnings():
        # we're setting the dask chunks
        warnings.filterwarnings("ignore", message="The specified chunks separate the stored", category=UserWarning)
        yield


class TestVIIRSJRRReader:
    """Test the VIIRS JRR L2 reader."""

    @pytest.mark.parametrize(
        "data_files",
        [
            lazy_fixture("surface_reflectance_file"),
            lazy_fixture("multiple_surface_reflectance_files"),
        ],
    )
    def test_get_dataset_surf_refl(self, data_files):
        """Test retrieval of datasets."""
        from satpy import Scene

        if not isinstance(data_files, list):
            data_files = [data_files]
        is_multiple = len(data_files) > 1
        bytes_in_m_row = 4 * 3200
        with set_chunk_size(bytes_in_m_row):
            scn = Scene(reader="viirs_edr", filenames=data_files)
            scn.load(["surf_refl_I01", "surf_refl_M01"])
        assert scn.start_time == START_TIME
        assert scn.end_time == END_TIME
        _check_surf_refl_data_arr(scn["surf_refl_I01"], multiple_files=is_multiple)
        _check_surf_refl_data_arr(scn["surf_refl_M01"], multiple_files=is_multiple)

    @pytest.mark.parametrize("filter_veg", [False, True])
    @pytest.mark.parametrize(
        "data_files",
        [
            lazy_fixture("surface_reflectance_with_veg_indices_file2"),
            lazy_fixture("multiple_surface_reflectance_files_with_veg_indices"),
        ],
    )
    def test_get_dataset_surf_refl_with_veg_idx(
            self,
            data_files,
            filter_veg,
    ):
        """Test retrieval of vegetation indices from surface reflectance files."""
        from satpy import Scene

        if not isinstance(data_files, list):
            data_files = [data_files]
        is_multiple = len(data_files) > 1
        bytes_in_m_row = 4 * 3200
        with set_chunk_size(bytes_in_m_row):
            scn = Scene(reader="viirs_edr", filenames=data_files,
                        reader_kwargs={"filter_veg": filter_veg})
            scn.load(["NDVI", "EVI", "surf_refl_qf1"])
        _check_vi_data_arr(scn["NDVI"], filter_veg, is_multiple)
        _check_vi_data_arr(scn["EVI"], filter_veg, is_multiple)
        _check_surf_refl_qf_data_arr(scn["surf_refl_qf1"], is_multiple)

    @pytest.mark.parametrize(
        ("var_names", "data_file"),
        [
            (("CldTopTemp", "CldTopHght", "CldTopPres"), lazy_fixture("cloud_height_file")),
            (("VLST",), lazy_fixture("lst_file")),
        ]
    )
    def test_get_dataset_generic(self, var_names, data_file):
        """Test datasets from cloud height files."""
        from satpy import Scene
        bytes_in_m_row = 4 * 3200
        with set_chunk_size(bytes_in_m_row):
            scn = Scene(reader="viirs_edr", filenames=[data_file])
            scn.load(var_names)
        for var_name in var_names:
            _check_continuous_data_arr(scn[var_name])

    def test_get_dataset_category(self, cloud_phase_file):
        """Test loading category (integer) data products."""
        from satpy import Scene
        bytes_in_m_row = 4 * 3200
        with set_chunk_size(bytes_in_m_row):
            scn = Scene(reader="viirs_edr", filenames=[cloud_phase_file])
            scn.load(["CloudPhase"])
        data_arr = scn["CloudPhase"]
        _array_checks(data_arr, dtype=np.int8)
        _shared_metadata_checks(data_arr)

    @pytest.mark.parametrize(
        ("aod_qc_filter", "exp_masked_pixel"),
        [
            (None, False),
            (0, True),
            (2, False)
        ],
    )
    def test_get_aod_filtered(self, aod_file, aod_qc_filter, exp_masked_pixel):
        """Test that the AOD product can be loaded and filtered."""
        from satpy import Scene
        bytes_in_m_row = 4 * 3200
        with set_chunk_size(bytes_in_m_row):
            scn = Scene(reader="viirs_edr", filenames=[aod_file], reader_kwargs={"aod_qc_filter": aod_qc_filter})
            scn.load(["AOD550"])
        _check_continuous_data_arr(scn["AOD550"])
        data_np = scn["AOD550"].data.compute()
        pixel_is_nan = np.isnan(data_np[-1, -1])
        assert pixel_is_nan if exp_masked_pixel else not pixel_is_nan

        # filtering should never affect geolocation
        lons, lats = scn["AOD550"].attrs["area"].get_lonlats()
        assert not np.isnan(lons[-1, -1].compute())
        assert not np.isnan(lats[-1, -1].compute())

    @pytest.mark.parametrize(
        ("data_file", "exp_available"),
        [
            (lazy_fixture("surface_reflectance_file"), False),
            (lazy_fixture("surface_reflectance_with_veg_indices_file"), True),
        ]
    )
    def test_availability_veg_idx(self, data_file, exp_available):
        """Test that vegetation indexes aren't available when they aren't present."""
        from satpy import Scene
        scn = Scene(reader="viirs_edr", filenames=[data_file])
        avail = scn.available_dataset_names()
        if exp_available:
            assert "NDVI" in avail
            assert "EVI" in avail
        else:
            assert "NDVI" not in avail
            assert "EVI" not in avail

    @pytest.mark.parametrize(
        ("filename_platform", "exp_shortname"),
        [
            ("npp", "Suomi-NPP"),
            ("JPSS-1", "NOAA-20"),
            ("J01", "NOAA-20"),
            ("n21", "NOAA-21")
        ])
    def test_get_platformname(self, surface_reflectance_file, filename_platform, exp_shortname):
        """Test finding start and end times of granules."""
        from satpy import Scene
        new_name = str(surface_reflectance_file).replace("npp", filename_platform)
        if new_name != str(surface_reflectance_file):
            shutil.copy(surface_reflectance_file, new_name)
        scn = Scene(reader="viirs_edr", filenames=[new_name])
        scn.load(["surf_refl_I01"])
        assert scn["surf_refl_I01"].attrs["platform_name"] == exp_shortname

    def test_volcanic_ash_drop_variables(self, volcanic_ash_file):
        """Test that Det_QF_Size variable is dropped when reading VolcanicAsh products.

        The said variable is also used as a dimension in v3r0 files, so the reading fails
        if it is not dropped.
        """
        from satpy import Scene
        scn = Scene(reader="viirs_edr", filenames=[volcanic_ash_file])
        available = scn.available_dataset_names()
        assert "Det_QF_Size" not in available


def _check_surf_refl_qf_data_arr(data_arr: xr.DataArray, multiple_files: bool) -> None:
    _array_checks(data_arr, dtype=np.uint8, multiple_files=multiple_files)
    _shared_metadata_checks(data_arr)
    assert data_arr.attrs["units"] == "1"
    assert data_arr.attrs["standard_name"] == "quality_flag"


def _check_vi_data_arr(data_arr: xr.DataArray, is_filtered: bool, multiple_files: bool) -> None:
    _array_checks(data_arr, multiple_files=multiple_files)
    _shared_metadata_checks(data_arr)
    assert data_arr.attrs["units"] == "1"
    assert data_arr.attrs["standard_name"] == "normalized_difference_vegetation_index"

    data = data_arr.data.compute()
    if is_filtered:
        np.testing.assert_allclose(data[0, :7], [np.nan, -1.0, -0.5, 0.0, 0.5, 1.0, np.nan])
        np.testing.assert_allclose(data[0, 8:8 + 16], np.nan)
        np.testing.assert_allclose(data[0, 8 + 16:], 0.0)
    else:
        np.testing.assert_allclose(data[0, :7], [np.nan, -1.0, -0.5, 0.0, 0.5, 1.0, np.nan])
        np.testing.assert_allclose(data[0, 8:], 0.0)


def _check_surf_refl_data_arr(
        data_arr: xr.DataArray,
        dtype: npt.DType = np.float32,
        multiple_files: bool = False
) -> None:
    _array_checks(data_arr, dtype, multiple_files=multiple_files)
    data = data_arr.data.compute()
    assert data.max() > 1.0  # random 0-1 test data multiplied by 100

    _shared_metadata_checks(data_arr)
    assert data_arr.attrs["units"] == "%"
    assert data_arr.attrs["standard_name"] == "surface_bidirectional_reflectance"


def _check_continuous_data_arr(data_arr: xr.DataArray) -> None:
    _array_checks(data_arr)

    if "valid_range" not in data_arr.attrs and "valid_min" not in data_arr.attrs:
        # random sample should be between 0 and 1 only if factor/offset applied
        exp_range = (0, 1)
    else:
        # if there is a valid range then we shouldn't be outside it
        exp_range = data_arr.attrs.get("valid_range",
                                       (data_arr.attrs.get("valid_min"), data_arr.attrs.get("valid_max")))
    data = data_arr.data.compute()
    assert not (data < exp_range[0]).any()
    assert not (data > exp_range[1]).any()

    _shared_metadata_checks(data_arr)


def _array_checks(data_arr: xr.DataArray, dtype: npt.Dtype = np.float32, multiple_files: bool = False) -> None:
    assert data_arr.dims == ("y", "x")
    assert isinstance(data_arr.attrs["area"], SwathDefinition)
    assert data_arr.attrs["area"].shape == data_arr.shape
    assert isinstance(data_arr.data, da.Array)
    assert np.issubdtype(data_arr.data.dtype, dtype)
    data_np = data_arr.data.compute()
    assert data_np.dtype == data_arr.dtype
    is_mband_res = _is_mband_res(data_arr)
    shape_multiplier = 1 + int(multiple_files)
    exp_shape = (M_ROWS * shape_multiplier, M_COLS) if is_mband_res else (I_ROWS * shape_multiplier, I_COLS)
    assert data_arr.shape == exp_shape
    exp_row_chunks = 4 if is_mband_res else 8
    assert all(c == exp_row_chunks for c in data_arr.chunks[0])
    assert data_arr.chunks[1] == (exp_shape[1],)


def _shared_metadata_checks(data_arr: xr.DataArray) -> None:
    is_mband_res = _is_mband_res(data_arr)
    exp_rps = 16 if is_mband_res else 32
    assert data_arr.attrs["sensor"] == "viirs"
    assert data_arr.attrs["rows_per_scan"] == exp_rps

    lons = data_arr.attrs["area"].lons
    lats = data_arr.attrs["area"].lats
    assert lons.attrs["rows_per_scan"] == exp_rps
    assert lats.attrs["rows_per_scan"] == exp_rps
    assert lons.min() >= -180.0
    assert lons.max() <= 180.0
    assert lats.min() >= -90.0
    assert lats.max() <= 90.0
    # Some files (ex. CloudHeight) have other lon/lats that shouldn't be used
    assert lons.attrs.get("long_name") != "BAD"
    assert lats.attrs.get("long_name") != "BAD"

    if "valid_range" in data_arr.attrs:
        valid_range = data_arr.attrs["valid_range"]
        assert isinstance(valid_range, tuple)
        assert len(valid_range) == 2

    if np.issubdtype(data_arr.dtype, np.floating):
        # floating point arrays always use NaN as fill and should not specify it in attrs
        assert "_FillValue" not in data_arr.attrs
    if "_FillValue" in data_arr.attrs:
        # make sure fill vlaue is scalar
        assert not hasattr(data_arr.attrs["_FillValue"], "shape")


def _is_mband_res(data_arr: xr.DataArray) -> bool:
    return "I" not in data_arr.attrs["name"]  # includes NDVI and EVI



# Copyright (c) 2025- Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittests for NWC SAF GEO HRW reader."""

import datetime as dt

import h5py
import numpy as np
import pytest

from satpy.readers.nwcsaf_hrw_nc import POST_V2025_DATASETS, PRE_V2025_SEVIRI_WIND_CHANNELS
from satpy.tests.utils import RANDOM_GEN

# This is the actual dtype of the trajectory items. We do not support it, so won't add this
# complexity to the test file creation. It's here anyway if someone wants to add it.
TRAJECTORY_DTYPE_ACTUAL = np.dtype([
    ("latitude", "<f4"),
    ("longitude", "<f4"),
    ("latitude_increment", "<f4"),
    ("longitude_increment", "<f4"),
    ("air_temperature", "<f4"),
    ("air_pressure", "<f4"),
    ("air_pressure_error", "<f4"),
    ("air_pressure_correction", "<f4"),
    ("wind_speed", "<f4"),
    ("wind_from_direction", "<f4"),
    ("quality_index_with_forecast", "i1"),
    ("quality_index_without_forecast", "i1"),
    ("quality_index_iwwg_value", "i1"),
    ("tracer_correlation_method", "i1"),
    ("tracer_type", "i1"),
    ("height_assignment_method", "i1"),
    ("orographic_index", "i1"),
    ("cloud_type", "i1"),
    ("correlation", "i1")
])

WIND_DTYPES = [
    ("wind_idx", np.dtype("uint32")),
    ("previous_wind_idx", np.dtype("uint32")),
    ("number_of_winds", np.dtype("uint8")),
    ("correlation_test", np.dtype("uint8")),
    ("quality_test", np.dtype("uint16")),
    ("segment_x", np.dtype("uint32")),
    ("segment_y", np.dtype("uint32")),
    ("segment_x_pix", np.dtype("uint32")),
    ("segment_y_pix", np.dtype("uint32")),
    ("latitude", np.dtype("<f4")),
    ("longitude", np.dtype("<f4")),
    ("latitude_increment", np.dtype("<f4")),
    ("longitude_increment", np.dtype("<f4")),
    ("air_temperature", np.dtype("<f4")),
    ("air_pressure", np.dtype("<f4")),
    ("air_pressure_error", np.dtype("<f4")),
    ("air_pressure_correction", np.dtype("<f4")),
    ("air_pressure_nwp_at_best_fit_level", np.dtype("<f4")),
    ("wind_speed", np.dtype("<f4")),
    ("wind_from_direction", np.dtype("<f4")),
    ("wind_speed_nwp_at_amv_level", np.dtype("<f4")),
    ("wind_from_direction_nwp_at_amv_level", np.dtype("<f4")),
    ("wind_speed_nwp_at_best_fit_level", np.dtype("<f4")),
    ("wind_from_direction_nwp_at_best_fit_level", np.dtype("<f4")),
    ("wind_speed_difference_nwp_at_amv_level", np.dtype("<f4")),
    ("wind_from_direction_difference_nwp_at_amv_level", np.dtype("<f4")),
    ("wind_speed_difference_nwp_at_best_fit_level", np.dtype("<f4")),
    ("wind_from_direction_difference_nwp_at_best_fit_level", np.dtype("<f4")),
    ("quality_index_with_forecast", np.dtype("uint8")),
    ("quality_index_without_forecast", np.dtype("uint8")),
    ("quality_index_iwwg_value", np.dtype("uint8")),
    ("tracer_correlation_method", np.dtype("uint8")),
    ("tracer_type", np.dtype("uint8")),
    ("height_assignment_method", np.dtype("uint8")),
    ("orographic_index", np.dtype("uint8")),
    ("cloud_type", np.dtype("uint8")),
    ("correlation", np.dtype("uint8")),
]

# Global attributes accessed by the file handler
NOMINAL_PRODUCT_TIME = np.bytes_("2025-02-06T13:00:00Z")
NOMINAL_PRODUCT_TIME_V2025 = np.bytes_("2025-09-23T13:00:00Z")
SPATIAL_RESOLUTION = np.array([3.], dtype=np.float32)
SATELLITE_IDENTIFIER = np.bytes_("MSG3")
SATELLITE_IDENTIFIER_V2025 = np.bytes_("MTI1")
ALGORITHM_VERSION_V2021 = np.bytes_("6.2.2")
ALGORITHM_VERSION_V2025 = np.bytes_("7.0")
SAMPLING_INTERVAL_V2025 = np.bytes_("10.00 minutes")

# Dataset attributes accessed by the file handler
TIME_PERIOD = np.array([15.], dtype=np.float32)

NUM_OBS = 123

FILENAME_INFO = {"platform_id": "MSG3", "region_id": "MSG-N-BS", "start_time": dt.datetime(2025, 2, 6, 13, 0)}
FILENAME_INFO_V2025 = {"platform_id": "MTI1", "region_id": "MSG-N-BS", "start_time": dt.datetime(2025, 9, 23, 13, 0)}
FILETYPE_INFO = {"file_type": "nc_nwcsaf_geo_hrw"}


@pytest.fixture(scope="module")
def hrw_file(tmp_path_factory):
    """Create a HRW data file."""
    fname = tmp_path_factory.mktemp("data") / "S_NWC_HRW_MSG3_MSG-N-BS_20250206T130000Z.nc"
    with h5py.File(fname, "w") as h5f:
        h5f.attrs["nominal_product_time"] = NOMINAL_PRODUCT_TIME
        h5f.attrs["spatial_resolution"] = SPATIAL_RESOLUTION
        h5f.attrs["satellite_identifier"] = SATELLITE_IDENTIFIER
        h5f.attrs["product_algorithm_version"] = ALGORITHM_VERSION_V2021
        for ch in PRE_V2025_SEVIRI_WIND_CHANNELS:
            _create_channel_dataset(h5f, ch)
    return fname


def _create_channel_dataset(h5f, name):
    dset = h5f.create_dataset(name, shape=(NUM_OBS,), dtype=np.dtype(WIND_DTYPES))
    dset.attrs["time_period"] = TIME_PERIOD
    data = _create_data()
    for i in range(NUM_OBS):
        dset[i] = data


def _create_data():
    data = []
    for key, dtype in WIND_DTYPES:
        val = np.array(255 * RANDOM_GEN.random(), dtype=dtype).item()
        data.append(val)

    return tuple(data)


@pytest.fixture(scope="module")
def hrw_v2025_file(tmp_path_factory):
    """Create a HRW data file in v2025 format."""
    fname = tmp_path_factory.mktemp("data") / "S_NWC_HRW_MTI1_MSG-N-BS_20250923T130000Z.nc"
    with h5py.File(fname, "w") as h5f:
        h5f.attrs["nominal_product_time"] = NOMINAL_PRODUCT_TIME_V2025
        h5f.attrs["spatial_resolution"] = SPATIAL_RESOLUTION
        h5f.attrs["satellite_identifier"] = SATELLITE_IDENTIFIER_V2025
        h5f.attrs["product_algorithm_version"] = ALGORITHM_VERSION_V2025
        h5f.attrs["sampling_interval"] = SAMPLING_INTERVAL_V2025
        # Just add one, they are pretty much the same:
        dset = h5f.create_dataset("wind_speed", shape=(NUM_OBS,), dtype=np.double)
        dset.attrs["standard_name"] = np.bytes_("wind_speed")
        dset.attrs["units"] = np.bytes_("m s-1")
        dset.attrs["valid_range"] = np.array([0., 409.4])
        dset.attrs["_FillValue"] = np.double(409.5)
        # Make the first value invalid
        data = np.array(409.4 * RANDOM_GEN.random((NUM_OBS,)), dtype=np.double)
        data[0] = 409.5
        dset[:] = data
        # And coordinates, those are needed
        lons = h5f.create_dataset("lon", shape=(NUM_OBS,), dtype=np.double)
        lons.attrs["standard_name"] = "longitude"
        lons.attrs["units"] = "degrees_east"
        lons.attrs["valid_range"] = np.array([-90., 90.], dtype=np.double)
        lons.attrs["_FillValue"] = np.double(245.)
        lons[:] = 180 * RANDOM_GEN.random((NUM_OBS,), dtype=np.double) - 90
        lats = h5f.create_dataset("lat", shape=(NUM_OBS,), dtype=np.double)
        lats.attrs["standard_name"] = "latitude"
        lats.attrs["units"] = "degrees_north"
        lats.attrs["valid_range"] = np.array([-180., 179.99999], dtype=np.double)
        lats.attrs["_FillValue"] = np.double(491.)
        lats[:] = 360 * RANDOM_GEN.random((NUM_OBS,), dtype=np.double) - 180.00001
    return fname



def test_hrw_handler_init(hrw_file):
    """Test the filehandler initialization works."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_file, FILENAME_INFO, FILETYPE_INFO)

    assert fh.filename_info == FILENAME_INFO
    assert fh.filetype_info == FILETYPE_INFO
    assert fh.platform_name is not None
    assert fh.sensor == "seviri"


def test_available_datasets(hrw_file):
    """Test that dynamic creation of available datasets works."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_file, FILENAME_INFO, FILETYPE_INFO)

    avail = fh.available_datasets()

    assert len(list(avail)) == len(PRE_V2025_SEVIRI_WIND_CHANNELS) * len(WIND_DTYPES)


def test_available_merged_datasets(hrw_file):
    """Test that dynamic creation of available datasets works when the channels are merge together."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_file, FILENAME_INFO, FILETYPE_INFO, merge_channels=True)

    avail = fh.available_datasets()

    assert len(list(avail)) == len(WIND_DTYPES)


def test_get_dataset(hrw_file):
    """Test reading a dataset."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_file, FILENAME_INFO, FILETYPE_INFO)

    data = fh.get_dataset({"name": "wind_vis06_wind_speed"}, {"the_answer": 42})

    assert data.attrs["the_answer"] == 42
    _check_common_attrs(data)


def test_get_merged_dataset(hrw_file):
    """Test reading a merged dataset."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_file, FILENAME_INFO, FILETYPE_INFO, merge_channels=True)

    data = fh.get_dataset({"name": "wind_speed"}, {})

    assert data.size == len(PRE_V2025_SEVIRI_WIND_CHANNELS) * NUM_OBS


def _check_common_attrs(data):
    assert "wind_vis06_longitude" in data.coords
    assert "wind_vis06_latitude" in data.coords
    assert data.dtype == np.float32
    assert data.values.dtype == np.float32
    assert "units" in data.attrs


def test_hrw_via_scene(hrw_file):
    """Test reading HRW datasets via Scene."""
    from satpy import Scene

    scn = Scene(reader="nwcsaf-geo", filenames=[hrw_file])
    scn.load(["wind_vis06_wind_from_direction"])
    data = scn["wind_vis06_wind_from_direction"]

    assert scn.start_time == dt.datetime(2025, 2, 6, 13, 0)
    assert scn.end_time == dt.datetime(2025, 2, 6, 13, 15)
    _check_scene_dataset_attrs(data)
    _check_common_attrs(data)


def _check_scene_dataset_attrs(data):
    from pyresample.geometry import SwathDefinition

    assert data.attrs["file_type"] == "nc_nwcsaf_geo_hrw"
    assert data.attrs["resolution"] == 3000.0
    assert data.attrs["name"] == "wind_vis06_wind_from_direction"
    assert data.attrs["coordinates"] == ("wind_vis06_longitude", "wind_vis06_latitude")
    assert data.attrs["start_time"] == dt.datetime(2025, 2, 6, 13, 0)
    assert data.attrs["end_time"] == dt.datetime(2025, 2, 6, 13, 15)
    assert data.attrs["reader"] == "nwcsaf-geo"
    assert isinstance(data.attrs["area"], SwathDefinition)
    assert data.attrs["area"].shape == (NUM_OBS,)


def test_merged_hrw_via_scene(hrw_file):
    """Test reading HRW merged datasets via Scene."""
    from satpy import Scene

    scn = Scene(reader="nwcsaf-geo", filenames=[hrw_file], reader_kwargs={"merge_channels": True})
    scn.load(["wind_from_direction"])
    data = scn["wind_from_direction"]

    assert data.shape == (len(PRE_V2025_SEVIRI_WIND_CHANNELS) * NUM_OBS,)


def test_hrw_handler_init_v2025(hrw_v2025_file):
    """Test the filehandler initialization works."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_v2025_file, FILENAME_INFO_V2025, FILETYPE_INFO)

    assert fh.filename_info == FILENAME_INFO_V2025
    assert fh.filetype_info == FILETYPE_INFO
    assert fh.platform_name is not None
    assert fh.sensor == "fci"


def test_available_datasets_v2025(hrw_v2025_file):
    """Test that dynamic creation of available datasets works."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_v2025_file, FILENAME_INFO, FILETYPE_INFO)

    avail = fh.available_datasets()
    avail_dsets = set(d[1]["name"] for d in avail)
    assert avail_dsets == set(POST_V2025_DATASETS)


def test_get_dataset_v2025(hrw_v2025_file):
    """Test reading a dataset from a file in v2025 format."""
    from satpy.readers.nwcsaf_hrw_nc import NWCSAFGEOHRWFileHandler

    fh = NWCSAFGEOHRWFileHandler(hrw_v2025_file, FILENAME_INFO, FILETYPE_INFO)

    data = fh.get_dataset({"name": "wind_speed"}, {"the_answer": 42})
    assert "units" in data.attrs
    assert data.shape == (NUM_OBS,)


def test_hrw_via_scene_v2025(hrw_v2025_file):
    """Test reading HRW v2025 datasets via Scene."""
    from satpy import Scene

    scn = Scene(reader="nwcsaf-geo", filenames=[hrw_v2025_file])
    scn.load(["wind_speed"])
    data = scn["wind_speed"]

    assert data.shape == (NUM_OBS,)
    assert "units" in data.attrs
    assert "standard_name" in data.attrs
    assert np.isnan(data[0].compute())


@pytest.mark.parametrize("dset", ["longitude", "latitude"])
def test_lon_and_lat_via_scene_v2025(hrw_v2025_file, dset):
    """Test reading longitude and latitude datasets v2025 datasets via Scene."""
    from satpy import Scene

    scn = Scene(reader="nwcsaf-geo", filenames=[hrw_v2025_file])
    scn.load([dset])
    data = scn[dset]

    assert data.shape == (NUM_OBS,)
    assert "units" in data.attrs
    assert "standard_name" in data.attrs
